#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 16 12:39:01 2022

    Zircon diffusion inverse problem
    Copyright (C) 2024 Jerome Fehrenbach

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

Solving Stefan problem in spherical geometry
(convention for V and W: oriented outwards)

This code solves the direct problem 
and proposes to estimate the temperature evolution from measurements of trace elements


Unnkown: temperature evolution Temp on a time interval [0,T]
- linear decrease of Temp : unknowns are Tini et Tfinal
- iterative refinement of Temp described by P1 finite elements (dyadic refinment) 

test case 1: 3 oscillations

@author: fehren
"""

import numpy as np
import matplotlib.pyplot as plt
import lib_stefan_trace as ls
import time as tm

# discretization parameters
nx = 500 # number of space points
nt = 500 # number of time steps
xi = 1-np.cos(np.pi*np.linspace(0,1,nx)/2)**2

#physical constants
tyear = 365*3600*24
tfin = 5000*tyear
dt = tfin/(nt-1)
t = np.linspace(0,tfin,nt)
tres = t/tfin # rescaled time

XH2O = 3.0

ZircNuc = 2e-4 # Zircon stable nuclei in cm,
U = 0 #zirconium undersaturation in the melt

Czirc = 490000.

# temperature profile
Td = 840+273
T0 = 720+273

Th = ls.TemperatureHistory(Td,T0,t,3,7)

# geometrical parameters
Xs = 1e-4 #zircon radius in cm, can be changed by the user
Rc = 0.1 # outer size of the domain

# scaling
Dscale = ls.DiffusionCoefficient((Th[0]+Th[-1])/2,3.0)
tscale = Rc**2/Dscale
dts = dt/tscale
time = t/tscale
ss = Xs/Rc
R = 1

cm=1/2.54
fig, ax = plt.subplots(figsize=(14*cm, 4*cm))
ax.plot(time*tscale/tyear,Th,'r',label='reference')
ax.set_xlabel('Time (years)',fontsize='x-large')
ax.set_ylabel('Temperature (K)',fontsize='x-large')
ax.set_xlim(0,5000)
plt.title('Temperature evolution',fontsize='xx-large')


V = 0. #V - zircon growth rate

constants = {'Czirc' : Czirc, 'Dscale' : Dscale, 'tscale' : tscale, 'tyear' : tyear}
constants['kpl']  = 0.1
constants['XH2O']  = 3.
constants['Rc']  = Rc

#%% generation of the synthetic data
utx1, tmp, VV1, tmp2, WW1, tmp3, Xstbl1, Rtbl1, CHstbl1, tmp4, tmp5 = ls.solve_direct_tangent(nt,nx,xi,dts,ss,R,Th,0*Th,constants)
ntr = np.shape(utx1)[2]-1

# additive noise in the observations
valeur = np.mean(CHstbl1,axis=0)
noiselevel = 0.04
noise = noiselevel*np.dot(np.random.randn(nt,ntr),np.diag(valeur))
CHstblobs = CHstbl1 + noise


species = ['Hf','Y','U','Th','Sm','Dy','Yb']
ch_max = np.max(CHstblobs,axis=0)
figw1, ax1 = plt.subplots()
for k in range(ntr):
    ax1.plot(Xstbl1[1:],CHstblobs[1:,k]/ch_max[k],'--',label=species[k])
ax1.set_xlabel('Position')
ax1.legend()
ax1.set_title('Trace elements (observations)')
### note :
### the observations are points: Xstbl1[1:], values: CHstblobs[1:,k] for species k
# The true temperature profile is Th

#%% parameter estimation
# cost function : F(beta) where beta are the parameters describing the temp. history
#               F(beta) = L(Xs_beta) u(beta) - u_obs
#       the observations are performed at the points x_obs_j, j=1..nobs
#       the concentration u(beta) is given at points Xs_beta, 
#           the operator L interpolates u(beta) at the observation points


# intial guess for the linear ramp
Tinik = Td-10
Tfinalk = T0+40
beta0 = [Tinik,Tfinalk]
B = np.zeros((nt,2)) # columns of B are the vectors of the basis of temperature evolution
B[:,0] = 1-tres
B[:,1] = tres

# initial profiles
Thini = np.dot(B,beta0)
ax.plot(time*tscale/tyear,Thini,'r:',label='ini.guess')
utxini, tmp, VV1, tmp2, WW1, tmp3, Xstblini, Rtbl1, CHstblini, tmp4, tmp5 = ls.solve_direct_tangent(nt,nx,xi,dts,ss,R,Thini,0*Th,constants)


betatbl = [beta0] # list of parameters (coefficient on the basis B)
Thktbl = [Thini] # list of temperature evolution
J = [] # list of values of J
gradJ = [] # list of values of ||grad J||

t1 = tm.time()
nbit = 20
# weight for the observation - normalized to the max observed value
value = np.max(CHstbl1,axis=0)
weight = np.zeros((nt,ntr))
for k in range(ntr):
    weight[:,k] = 1/value[k]
weightr = np.reshape(weight,nt*ntr)

betak = beta0
jj = 0 # depth of refinment: time points q/2**jj
for nit in range(nbit):
    print('iteration ',nit,'/',nbit)
    Thk = np.dot(B,betak)
    # assembly of the Jacobian
    nbparam = np.shape(B)[1]
    DG = np.zeros((nt*ntr,nbparam))
    for j in range(nbparam):
        alphaj = B[:,j]
        Jobs, Gk, DGalphaj = ls.compute_G_DG(nt,nx,xi,dts,ss,R,Thk,alphaj,constants,Xstbl1,CHstblobs)
        DG[:,j] = weightr*np.reshape(DGalphaj,ntr*nt)
        G = weightr*np.reshape(Gk,ntr*nt)
    A = np.dot(DG.transpose(),DG)
    s = np.linalg.eigvalsh(np.dot(DG.transpose(),DG))+1e-7
    sm = np.dot(DG.transpose(),G)
    J.append(np.linalg.norm(G))
    gradJ.append(np.linalg.norm(sm))
    if nit==0:
        ngrad0 = np.linalg.norm(sm)
    if np.linalg.norm(sm)<1e-5*ngrad0:
        break
    direction = np.linalg.solve(A+1e-3*np.max(s)*np.eye(nbparam),-sm)
    betak += direction
    betatbl.append(betak)
    Thktbl.append(np.dot(B,betak))
    if np.mod(nit,4)==3:
        # refinment
        jj += 1
        ncols = np.shape(B)[1]
        Bnew = np.zeros((nt,ncols+2**(jj-1)))
        Bnew[:,:ncols] = B
        for q in range(2**(jj-1)):
            tent = np.minimum(2**jj*tres-2*q,(2*q+2)-2**jj*tres)
            ei = np.maximum(tent,0)
            Bnew[:,ncols+q] = ei
        B = Bnew
        betanew = np.zeros(ncols+2**(jj-1))
        betanew[:ncols] = betak
        betak = betanew
t2 = tm.time()
    
print('Computation time (s):',np.round(t2-t1,3))


  
plt.figure()
plt.plot(gradJ)
plt.title('Grad of cost function')

plt.figure()
plt.semilogy(gradJ)
plt.title('Grad of cost function - log scale')

plt.figure()
plt.plot(J)
plt.title('Cost function')

# solve woth the final temperature profile obtained
utxfinal, tmp, VV1, tmp2, WW1, tmp3, Xstblfinal, Rtbl1, CHstblfinal, tmp4, tmp5 = ls.solve_direct_tangent(nt,nx,xi,dts,ss,R,Thk,0*Th,constants)


ch_max = np.max(CHstblfinal,axis=0)
fig2, ax2 = plt.subplots(figsize=(14*cm, 4*cm))
ax2.set_prop_cycle(color=['c', 'm', 'y', 'k','r','g','b'])
for k in range(ntr):
    ax2.plot(Xstblini[1:],CHstblini[1:,k]/ch_max[k],':')
for k in range(ntr):
    ax2.plot(Xstblfinal[1:],CHstblobs[1:,k]/ch_max[k],'--')
for k in range(ntr):
    ax2.plot(Xstblfinal[1:],CHstblfinal[1:,k]/ch_max[k],label=species[k])
ax2.set_xlabel('Position (microns)',fontsize='x-large')
ax2.set_ylabel('Relative concentration',fontsize='x-large')
ax2.set_xlim(0,np.max(Xstblfinal[1:]))
ax2.legend(fontsize='x-large')
ax2.set_title('Trace elements before/after optimization',fontsize='xx-large')
ax2.tick_params(labelsize='large')

for k,th in enumerate(Thktbl[:-2]):
    if np.mod(k,4)==3:
        ax.plot(time*tscale/tyear,th,label='iteration'+str(k),linewidth=0.5)
ax.plot(time*tscale/tyear,Thktbl[-1],'k',label='iteration'+str(nbit-1),linewidth=2)
ax.legend(fontsize='x-large')
ax.set_xlim(0,tfin/tyear)
ax.tick_params(labelsize='large')
