#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Nov 16 12:39:01 2022

    Zircon diffusion inverse problem
    Copyright (C) 2024 Jerome Fehrenbach

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

Solving Stefan problem in spherical geometry
(convention for V and W: oriented outwards)

This code solves the direct problem 
and proposes to estimate the temperature history from measurements of trace elements


Unnkown : temperature history Temp on a time interval [0,T], and the total time T
- step 1 linear decrease of Temp : unknowns are Tini and T
when T is known:
- iterative refinement of Temp described by P1 finite elements (dyadic refinment) 

test case 3 : real data Fish Canyon


@author: fehren
"""

import numpy as np
import matplotlib.pyplot as plt
import lib_stefan_trace_4elts as ls
import time as tm
import pandas as pd

# data loading and plotting
df = pd.read_csv('./Trace-elements_long.csv',sep=',')
print(df.keys())
xtbl = np.array(df['Distance'])

st_tr = 4 # starting index for valid trace obs

chsobs = np.array(df[['Hf','Y','U','Th']]) #  

fig, ax = plt.subplots()
ax.plot(xtbl,chsobs,'o')

#%%
# discretization parameters
nx = 200 # number of space points
nt = 200 # number of time steps
xi = 1-np.cos(np.pi*np.linspace(0,1,nx)/2)**2

#physical constants
tyear = 365*3600*24
tfin = 4000*tyear
dt = tfin/(nt-1)
t = np.linspace(0,tfin,nt)
tres = t/tfin # rescaled time

ZircNuc = 2e-4 # Zircon stable nuclei in cm,
U = 0 #zirconium undersaturation in the melt

Czirc = 490000.

# temperature profile
Td = 815+273
T0 = 745+273

Th = ls.TemperatureHistory(Td,T0,t,3,0)


# geometrical parameters
Xs = xtbl[st_tr]/2e4 #zircon radius in cm, can be changed by the user
Rc = 0.3 # outer size of the domain

Mas0 = 4*np.pi*Xs**3*Czirc/3+4*np.pi*ls.ZrSaturation(Td)*(Rc**3-Xs**3)/3
Mpl=0

# scaling
Dscale = ls.DiffusionCoefficient((Th[0]+Th[-1])/2,5.0)
tscale = Rc**2/Dscale
dts = dt/tscale
time = t/tscale
ss = Xs/Rc
R = 1


fig, ax = plt.subplots()
ax.set_xlabel('Time (years)')
plt.title('Temperature profile')

#%
V = 0. #V - zircon growth rate

constants = {'Czirc' : Czirc, 'Dscale' : Dscale, 'tscale' : tscale, 'tyear' : tyear}
constants['kpl']  = 0.0
constants['XH2O']  = 5.
constants['Rc']  = Rc


VV = np.zeros(nt)
tt = np.zeros(nt)


CHstblobs = chsobs[st_tr:,:]*1000
average = np.mean(CHstblobs[:17,:],axis=0)/ls.Kdd_5elts(Td)
Xstbl1 = xtbl[st_tr:]


species = ['Hf','Y','U','Th']
ntr = 4

ch_max = np.max(CHstblobs,axis=0)
figw1, ax1 = plt.subplots()
for k in range(ntr):
    ax1.plot(Xstbl1[1:],CHstblobs[1:,k]/ch_max[k],'--',label=species[k])
ax1.set_xlabel('Position')
ax1.legend()
ax1.set_title('Trace elements (observations)')
### note :
### the observations are points: Xstbl1[1:], values: CHstblobs[1:,k] for species k
#
#%% parameter estimation step 1: linear ramp with unknowns Tini and T (dt)
# cost function : F(beta) where beta are the parameters describing the temp. history
#               F(beta) = L(Xs_beta) u(beta) - u_obs
#       the observations are performed at the points x_obs_j, j=1..nobs
#       the concentration u(beta) is given at points Xs_beta, 
#           the operator L interpolates u(beta) at the observation points


# intial guess for the linear ramp
Tinik = Td
Tfinal = T0
dtk = dts
tfink = (nt-1)*dtk*tscale
tk = np.linspace(0,tfink,nt)
timek = tk/tscale

beta0 = [Tinik,np.log(dtk)]
B = np.zeros((nt,1))
B[:,0] = 1-tres

# initial profiles
Thini = Tinik*(1-tres) + Tfinal*tres
ax.plot(timek*tscale/tyear,Thini,'r:',label='ini.guess')
ax.legend()
utxini, tmp, VV1, tmp2, WW1, tmp3, Xstblini, Rtbl1, CHstblini, tmp4, tmp5 = ls.solve_direct_tangent(nt,nx,xi,dtk,ss,R,Thini,0*Th,constants,average)


betatbl = [beta0] # list of the parameters
Thktbl = [Thini] # list of temperature evolution
timektbl = [timek] # list of total time
tinitbl=[Tinik] # list of initial temperature
dttbl=[dtk] # list of time step

J = [] # list of cost function
gradJ = [] # list of gradient of cost function
t1 = tm.time()
nbit1 = 10

# weight for the observation - normalized to the median observed value
valeur = np.median(CHstblobs,axis=0)
weight = np.zeros((nt,ntr))
for k in range(ntr):
    weight[:,k] = 1/valeur[k]
#weight[:,0] = 0 # no account for Hf 
    
weightr = np.reshape(weight,nt*ntr)

betak = beta0
Thk = Tinik*(1-tres) + Tfinal*tres

for nit in range(nbit1):
    print('iteration ',nit,'/',nbit1,'Tini:',Tinik,'dt:',dtk)
    Thk = Tinik*(1-tres) + Tfinal*tres
    # assembly of the Jacobian
    DG = np.zeros((nt*ntr,2))
    Jobs, Gkk, DGtini, finalradius, Dfinalradius = ls.compute_G_DG(nt,nx,xi,dtk,ss,R,Thk,B[:,0],constants,Xstbl1,CHstblobs,average)
    DG[:nt*ntr,0] = weightr*np.reshape(DGtini,ntr*nt)
    
    Jobs, Gk, DGtini, finalradius, Dfinalradius = ls.compute_G_DG(nt,nx,xi,dtk,ss,R,Thk,1-B[:,0],constants,Xstbl1,CHstblobs,average)
    DG[:nt*ntr,1] = weightr*np.reshape(DGtini,ntr*nt)
    print('final radius',finalradius)
    
    G = weightr*np.reshape(Gk,ntr*nt)
    A = np.dot(DG.transpose(),DG)
    s = np.linalg.eigvalsh(A)+1e-7
    sm = np.dot(DG.transpose(),G)
    J.append(np.linalg.norm(G))
    gradJ.append(np.linalg.norm(sm))
    if nit==0:
        ngrad0 = np.linalg.norm(sm)
    if np.linalg.norm(sm)<1e-5*ngrad0:
        break
    direction = np.linalg.solve(A+1e-2*np.max(s)*np.eye(2),-sm)
    betak += direction
    Tinik += direction[0]
    dtk = dtk/(1-direction[1]/(Tinik-Tfinal))
    tfink = (nt-1)*dtk*tscale
    tk = np.linspace(0,tfink,nt)
    timek = tk/tscale
    
    betatbl.append(betak.copy())
    tinitbl.append(Tinik)
    dttbl.append(dtk)
    Thktbl.append(Tinik*(1-tres) + Tfinal*tres)
    timektbl.append(timek)
    
t2 = tm.time()
    
print('Computation time (s)',np.round(t2-t1,3))

  
plt.figure()
plt.plot(gradJ)
plt.title('Grad of cost function')

plt.figure()
plt.semilogy(gradJ)
plt.title('Grad of cost function - log scale')

plt.figure()
plt.plot(J)
plt.title('Cost function')



utxfinal, tmp, VV1, tmp2, WW1, tmp3, Xstblfinal, Rtbl1, CHstblfinal, tmp4, tmp5 = ls.solve_direct_tangent(nt,nx,xi,dtk,ss,R,Thk,0*Th,constants,average)


ch_max = np.max(CHstblobs,axis=0)
fig2, ax2 = plt.subplots()
ax2.set_prop_cycle(color=['c', 'm', 'y', 'k'])
for k in range(ntr-1):
    ax2.plot(Xstbl1[1:],CHstblobs[1:,k]/ch_max[k],':',label='obs.')
for k in range(ntr-1):
    ax2.plot(Xstblini[1:],CHstblini[1:,k]/ch_max[k],'--',label='ini.')
for k in range(ntr-1):
    ax2.plot(Xstblfinal[1:],CHstblfinal[1:,k]/ch_max[k],label=species[k]+' after')
ax2.set_xlabel('Position')
ax2.legend()
xmini=np.min(Xstbl1)
xmaxi=np.max(Xstbl1)
ax2.set_ylim(0,2)
ax2.set_xlim(xmini,xmaxi)
ax2.set_title('Trace elements after optimization')

for k in range(ntr-1):
    figk,axk = plt.subplots()
    axk.plot(Xstblini[1:],CHstblini[1:,k],label=species[k]+' ini')
    axk.plot(Xstblfinal[1:],CHstblfinal[1:,k],label=species[k]+' after')
    axk.plot(Xstbl1[1:],CHstblobs[1:,k],'o:',label='obs.')
    axk.set_title(species[k])
    axk.legend()

for k,th in enumerate(Thktbl[:-1]):
    if np.mod(k,4)==2:
        ax.plot(timektbl[k]*tscale/tyear,th,label='ramp iteration '+str(k),linewidth=0.5)
ax.plot(timektbl[-1]*tscale/tyear,Thktbl[-1],'k', dashes=[6, 2],label='best ramp',linewidth=2)


#%%############ coarse to fine approach
Thktbl = [] # list of temperature evolution
J = [] # list of cost function
gradJ = [] # list of gradient of cost function
t1 = tm.time()
nbit = 16
valeur = np.max(CHstblobs,axis=0)

betak = np.array([Tinik,Tfinal])
jj = 0 # depth of refinment : time points q/2**jj
B = np.zeros((nt,2))
B[:,0] = 1-tres
B[:,1] = tres

cm = 1/2.54
fig3,ax3 = plt.subplots(4,3,figsize=(10*cm, 15*cm))

for nit in range(nbit):
    print('iteration ',nit,'/',nbit)
    Thk = np.dot(B,betak)
    # assembly of the Jacobian
    nbparam = np.shape(B)[1]
    DG = np.zeros((nt*ntr,nbparam))
    for j in range(nbparam):
        alphaj = B[:,j]
        Jobs, Gk, DGalphaj, finalradius, Dfinalradius = ls.compute_G_DG(nt,nx,xi,dtk,ss,R,Thk,alphaj,constants,Xstbl1,CHstblobs,average)
        DG[:,j] = weightr*np.reshape(DGalphaj,ntr*nt)
    G = weightr*np.reshape(Gk,ntr*nt)
    A = np.dot(DG.transpose(),DG)
    s = np.linalg.eigvalsh(np.dot(DG.transpose(),DG))+1e-7
    sm = np.dot(DG.transpose(),G)
    J.append(np.linalg.norm(G))
    gradJ.append(np.linalg.norm(sm))
    if nit==0:
        ngrad0 = np.linalg.norm(sm)
    if np.linalg.norm(sm)<1e-5*ngrad0:
        print('converged',nit)
        break
    direction = np.linalg.solve(A+1e-3*np.max(s)*np.eye(nbparam),-sm)
    betak += direction
    betatbl.append(betak)
    Thktbl.append(np.dot(B,betak))
    if np.mod(nit,4)==0:
        # refinment
        jj += 1
        ncols = np.shape(B)[1]
        Bnew = np.zeros((nt,ncols+2**(jj-1)))
        Bnew[:,:ncols] = B
        for q in range(2**(jj-1)):
            tent = np.minimum(2**jj*tres-2*q,(2*q+2)-2**jj*tres)
            ei = np.maximum(tent,0)
            Bnew[:,ncols+q] = ei
        B = Bnew
        betanew = np.zeros(ncols+2**(jj-1))
        betanew[:ncols] = betak
        betak = betanew
    bb = -1
    if nit==7:
        bb = 0
    if nit==11:
        bb = 1
    if nit==15:
        bb=2
    
    for k in range(ntr):
        if bb>=0:
            tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, Xstblfinal, tmp6, CHstblfinal, tmp7, tmp8 = ls.solve_direct_tangent(nt,nx,xi,dtk,ss,R,Thk,0*Th,constants,average)

            ax3[k,bb].plot(Xstblini[1:],CHstblini[1:,k],label=species[k]+' best ramp')
            ax3[k,bb].plot(Xstblfinal[1:],CHstblfinal[1:,k],label=species[k]+' '+str(nit)+' it.')
            ax3[k,bb].plot(Xstbl1[1:],CHstblobs[1:,k],'o:',label='obs.')
            ax3[k,bb].set_xlabel('Position (microns)')
            ax3[k,bb].set_ylabel('Concentration (ppm)')
            ax3[k,bb].set_xlim(np.min(Xstblini[1:]),np.max(Xstblini[1:]))
            if k==1:
                ax3[k,bb].set_ylim(10,250)
            ax3[k,bb].legend()

t2 = tm.time()
print('Computation time (s):',np.round(t2-t1,3))


Thk = np.dot(B,betak)

plt.figure()
plt.plot(gradJ)
plt.title('Grad of cost function')

plt.figure()
plt.semilogy(gradJ)
plt.title('Grad of cost function - log scale')

plt.figure()
plt.plot(J)
plt.title('Cost function')

utxfinal, tmp, VV1, tmp2, WW1, tmp3, Xstblfinal, Rtbl1, CHstblfinal, tmp4, tmp5 = ls.solve_direct_tangent(nt,nx,xi,dtk,ss,R,Thk,0*Th,constants,average)

ch_max = np.max(CHstblobs,axis=0)
fig2, ax2 = plt.subplots()
ax2.set_prop_cycle(color=['c', 'm', 'y', 'k'])
for k in range(ntr):
    ax2.plot(Xstbl1[1:],CHstblobs[1:,k]/ch_max[k],':',label='obs.')
for k in range(ntr):
    ax2.plot(Xstblini[1:],CHstblini[1:,k]/ch_max[k],'--',label='ini.')
for k in range(ntr):
    ax2.plot(Xstblfinal[1:],CHstblfinal[1:,k]/ch_max[k],label=species[k]+' after')
ax2.set_xlabel('Position')
ax2.legend()
xmini=np.min(Xstbl1)
xmaxi=np.max(Xstbl1)
ax2.set_ylim(0,2)
ax2.set_xlim(xmini,xmaxi)
ax2.set_title('Trace elements after optimization')


for k,th in enumerate(Thktbl[:-1]):
    if np.mod(k,4)==3:
        ax.plot(timektbl[-1]*tscale/tyear,th,label='iteration'+str(k),linewidth=0.5)
ax.plot(timektbl[-1]*tscale/tyear,Thktbl[-1],'k',label='iteration'+str(nbit-1),linewidth=2)
ax.legend()
fig.suptitle('nb of iterations: '+str(nbit))
 
