#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan  3 17:30:41 2023

    Zircon diffusion inverse problem
    Copyright (C) 2024 Jerome Fehrenbach

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

library for the Stefan problem 
direct and tangent solve

includes the trace elements
uses cython to improve speed

@author: fehrenbach
"""

import solver_cython as sc
import numpy as np
import matplotlib.pyplot as plt


def ZrSaturation(T): #defining Zr saturation conditions
    Mfactor = 0.0000048*T**2-0.0083626*T+4.8484463; # empirical relations from magma
    Csat = 490000/np.exp(10108/T-1.16*(Mfactor-1)-1.48) # Boehnke et al 2013 ChemGeol 351,324 assuming 490,000 ppm in Zircon
    return Csat

def DZrSaturation(T): #derivative of Zr saturation conditions
    Mfactor = 0.0000048*T**2-0.0083626*T+4.8484463 # empirical relations from magma
    DMfactor = 2*0.0000048*T-0.0083626
    DCsat = 490000*(10108/T**2 + 1.16*DMfactor)*np.exp(-10108/T+1.16*(Mfactor-1)+1.48)
    return DCsat

def TemperatureHistory(Td,T0,time,n=0,amp=0):
    # define an oscillating and decreasing temperature profile
    #   sum of a linear ramp+oscillating sin
    # IN : Td/T0 : initial/final temperature for the ramp term
    #       time : discretization time steps
    #       n : number of oscillations in the interval
    #       amp : amplitude of the oscillations
    # OUT : temperature history
    reltime = time/time[-1]
    Th = np.maximum(T0,Td-(Td-T0)*reltime-amp*np.sin(n*2*np.pi*reltime))
    return Th


def DiffusionCoefficient(T,x = 3):  
    #defining Zr diffusion coefficients in melts as f(T,XH2O)
    # IN : T : temperature
    #       x : percentage of H2O content
    #best fit of Zr Diff coefficients (several workers) and WH83 dependence on XH2O
    theta = 1000./T
    lnD = -(11.4*x+3.13)/(0.84*x+1)-(21.4*x+47)/(1.06*x+1)*theta
    dif = np.exp(lnD)*1e4 # in cm2/s
    return dif


def DDiffusionCoefficient(T,x = 3):  
    #derivative wrt T of Zr diffusion coefficient in melts as f(T,XH2O)
    theta= 1000./T
    lnD=-(11.4*x+3.13)/(0.84*x+1)-(21.4*x+47)/(1.06*x+1)*theta
    fact = (21.4*x+47)/(1.06*x+1)*1000/T**2
    Ddif = fact*np.exp(lnD)*1e4 # in cm2/s/K
    return Ddif


def TraceDiff(T, x=3):
    # Tr diffusion coefficient in melts as f(T,XH2O)
    # IN : T : temperature
    #       x : percentage of H2O content
    # OUT : diffusion coef for the trace elements
    #       ordering : Hf, Y, U, Th, Sm, Dy, Yb
    lnD_Hf = np.log(DiffusionCoefficient(T,x)/3e4)
    lnD_Y = -9.925340370-(35587.24428-2615.214492*x)/T
    P = 0.01 #Pressure in GPa
    lnD_U = -6.37-2.65*np.sqrt(x) - (44729-1093*P-8944*np.sqrt(x))/T
    lnD_Th = -7.02-1.30*np.sqrt(x) - (44682-1370*P-7281*np.sqrt(x))/T
    Dif = 1e4*np.exp([lnD_Hf,lnD_Y,lnD_U,lnD_Th,lnD_Y,lnD_Y,lnD_Y])
    return Dif


def DTraceDiff(T, x=3):
    # temperature derivative of Tr diffusion coefficient in melts as f(T,XH2O)
    # IN : T : temperature
    #       x : percentage of H2O content
    # OUT : derivative of diffusion coef for the trace elements
    #       ordering : Hf, Y, U, Th, Sm, Dy, Yb
    lnD_Hf = np.log(DiffusionCoefficient(T,x)/3e4)
    DHf = DDiffusionCoefficient(T,x)/DiffusionCoefficient(T,x)
    lnD_Y = -9.925340370-(35587.24428-2615.214492*x)/T
    DY = (35587.24428-2615.214492*x)/T**2
    P = 0.01 #Pressure in GPa
    lnD_U = -6.37-2.65*np.sqrt(x) - (44729-1093*P-8944*np.sqrt(x))/T
    DU = (44729-1093*P-8944*np.sqrt(x))/T**2
    lnD_Th = -7.02-1.30*np.sqrt(x) - (44682-1370*P-7281*np.sqrt(x))/T
    DTh = (44682-1370*P-7281*np.sqrt(x))/T**2
    fact = np.array([DHf,DY,DU,DTh,DY,DY,DY])
    DDif = fact*1e4*np.exp([lnD_Hf,lnD_Y,lnD_U,lnD_Th,lnD_Y,lnD_Y,lnD_Y])
    return DDif



def Kdd(T):
    # partitioning coef for the trace elements, as a function of temperature
    # IN : T : temperature
    # OUT : kdd : vector containing the partition coef
    #       ordering : Hf, Y, U, Th, Sm, Dy, Yb
    X=1000/T;
    KD_Hf = 11.29 * X - 2.275
    KD_Y = 19.47 * X - 13.04
    KD_U = 15.32 * X - 9.17
    KD_Th = 13.02 * X - 8.354
    Csat = ZrSaturation(T)
    KD_Sm = np.log(13.338*Csat**(-0.622))
    KD_Dy = np.log(2460.0*Csat**(-0.867))
    KD_Yb = np.log(33460*Csat**(-1.040))
    kdd = np.exp(np.array([KD_Hf,KD_Y,KD_U,KD_Th,KD_Sm,KD_Dy,KD_Yb]))
    return kdd

def DKdd(T):
    # derivative of the partitioning coef for the trace elements wrt temperature
    # IN : T : temperature
    # OUT : Dkdd : vector containing the derivative of the partition coef
    #       ordering : Hf, Y, U, Th, Sm, Dy, Yb
    X=1000/T;
    KD_Hf = 11.29 * X - 2.275
    KD_Y = 19.47 * X - 13.04
    KD_U = 15.32 * X - 9.17
    KD_Th = 13.02 * X - 8.354
    Csat = ZrSaturation(T)
    DCsat = DZrSaturation(T)
    KD_Sm = np.log(13.338*Csat**(-0.622))
    KD_Dy = np.log(2460.0*Csat**(-0.867))
    KD_Yb = np.log(33460*Csat**(-1.040))
    fSm = -0.622*DCsat/Csat
    fDy = -0.867*DCsat/Csat
    fYb = -1.040*DCsat/Csat
    fact = [-11.29*1000/T**2, -19.47*1000/T**2, -15.32*1000/T**2, -13.02*1000/T**2,fSm,fDy,fYb]
    Dkdd = fact*np.exp(np.array([KD_Hf,KD_Y,KD_U,KD_Th,KD_Sm,KD_Dy,KD_Yb]))
    return Dkdd


def Xplag(T):
    x=T-273;
    Xpl = 1.94319815919224E-07*x**4+-6.44583169395759E-04*x**3+0.799105630147523*x**2+-439.159560503417*x+90385.6027471822
    Xpl=Xpl/100
    return Xpl

def DXplag(T):
    x=T-273;
    DXpl = 1.94319815919224E-07*4*x**3+-6.44583169395759E-04*3*x**2+0.799105630147523*x*2+-439.159560503417
    DXpl=DXpl/100
    return DXpl

def D2Xplag(T):
    x=T-273;
    D2Xpl = 1.94319815919224E-07*4*3*x**2+-6.44583169395759E-04*3*2*x+0.799105630147523*2
    D2Xpl=D2Xpl/100
    return D2Xpl





def compute_G_DG(nt,nx,xi,dts,Xsini,Rini,Th,alpha,constants,Xsobs,CHstblobs):
    # compute G = the difference between the predictions and the obs.
    #       and DG.alpha the derivative of G in the direction alpha
    # 
    # IN : nt, nx : number of time steps/space steps
    #       xi : location of space points
    #       dts : value of the (scaled) time step
    #       Xsini : initial position of the crystal boundary
    #       Rini : initial outer radius of the domain
    #       Th : temperature evolution
    #       alpha : direction of variation of Th
    #       constants : a dict containing the constant values : 
    #           Rc - intial outer radius (length scale)
    #           Dscale - diffusion scale
    #           tscale - time scale
    #           tyear - duration of 1 year in seconds
    #           Czirc - concentration of ZR in crystal
    #           XH2O - content in H2O in %
    #       Xsobs : value of Xs for the observations
    #       CHstblobs : value of observed trace elements
    # OUT : Jobs, G, DG.alpha
    #       Jobs : locations of valid obs points
    #       G, DGalpha : value of G and derivative in the direction alpha
    #
    # The observations are the values of the concentrations of trace elts
    #       at the boundary of the crystal (Xs[:],CHstbl[:]) for all t
    # G and DG are obtained by interpolating these observations on the
    #       observations points Xsobs. For this it is necessary to set to 0
    #       the values outside the range of (Xs[:])
    #       The set of valid indices for Xsobs is denoted Jobs
    #       The size of G is ntr x nt : ntr=number of trace elts 
    #       
    ntr = np.shape(CHstblobs)[1] # number of trace elements
    
    utx, vtx, VV, DVV, WW, DWW, Xstbl, Rtbl, CHstbl, DXs, DCHstbl = solve_direct_tangent(nt,nx,xi,dts,Xsini,Rini,Th,alpha,constants)
    Jobs = np.where(np.logical_and(Xsobs>np.min(Xstbl),Xsobs<np.max(Xstbl)))[0]

    G = np.zeros((nt,ntr))
    DGalpha = np.zeros((nt,ntr))
    tt = np.zeros((nt,1))
    Dtt = np.zeros((nt,1))
    nxx = np.shape(Xstbl)[0]
    for k,j in enumerate(Jobs):
        i = np.max(np.where(Xsobs[j]>=Xstbl))
        ip1 = np.min([i+1,nxx-1])
        t = (Xstbl[ip1]-Xsobs[j])/(Xstbl[ip1]-Xstbl[i])
        if np.isinf(t):
            t = 0.5
        tt[j] = t
        G[j,:] = t*CHstbl[i,:] + (1-t)*CHstbl[ip1,:] - CHstblobs[j,:]
        dt = (DXs[ip1]*(Xstbl[ip1]-Xstbl[i])-(Xstbl[ip1]-Xsobs[j])*(DXs[ip1]-DXs[i]))/(Xstbl[ip1]-Xstbl[i])**2
        if np.isnan(dt):
            dt = 0
        Dtt[j] = dt
        DLCHobs = dt*CHstbl[i,:] -dt*CHstbl[ip1,:] 
        DGalpha[j,:] = t*DCHstbl[i,:] + (1-t)*DCHstbl[ip1,:] + DLCHobs

    return Jobs, G, DGalpha
    


def solve_direct_tangent(nt,nx,xi,dts,Xsini,Rini,Th,alpha,constants,affiche=False):
    # solve the direct problem
    # using implicit time discretization
    #
    # The direct and tangent model are solved as fixed point pbs with explicit
    #   functions to be iterated
    #
    # IN : nt, nx : number of time steps/space steps
    #       xi : location of space points
    #       dts : value of the (scaled) time step
    #       Xsini : initial position of the crystal boundary
    #       Rini : initial outer radius of the domain
    #       Th : temperature history
    #       alpha : direction of variation of Th
    #       constants : a dict containing the constant values : 
    #           Rc - intial outer radius (length scale)
    #           Dscale - diffusion scale
    #           tscale - time scale
    #           tyear - duration of 1 year in seconds
    #           Czirc - concentration of ZR in crystal
    #           XH2O - content in H2O in %
    #       affiche = boolean that indicates display/no display
    # OUT : utx : value of u(t,x,k) = concentration of species k at (t,x)
    #               for k=0 : Zr, other species for k>=1
    #       vtx : derivative of utx in the direction alpha
    #       VV/DVV : value/derivative of the velocity V along time
    #       WW/DWW : value of the velocity W along time
    #       Xstbl /Rtbl/CHstbl: value of Xs (crystal radius)/R (outer radius)/Trace elts 
    #                           concentration along time
    #       DXstbl /DRtbl/DCHstbl: derivatives of Xs (crystal radius)/R (outer radius)/Trace elts 
    #                           concentration along time
    # note : for Xs, CHstbl, DXs, DCHstbl only valid time steps are provided (when melt is still present)

    ntr = 7 # number of trace elements
    c0 = ZrSaturation(Th[0])
    Dc0 = DZrSaturation(Th[0])*alpha[0]
    Rc = constants['Rc']
    Dscale = constants['Dscale']
    tscale = constants['tscale']
    tyear = constants['tyear']
    Czirc = constants['Czirc']
    xh2o = constants['XH2O']
    
    utx = np.zeros((nt,nx,ntr+1))
    vtx = np.zeros((nt,nx,ntr+1))
    VV = np.zeros(nt)
    WW = np.zeros(nt)
    DVV = np.zeros(nt)
    DWW = np.zeros(nt) 
    DR = np.zeros(nt)
    DXs = np.zeros(nt)
    
    Mpl = 0
    Mas0 = 4*np.pi*(Rc*Xsini)**3*Czirc/3 + 4*np.pi*c0*(Rc**3-Xsini**3)/3

    ### Zr initial condition for t = 0
    u0 = np.zeros(nx)
    u0[:] = c0
    utx[0,:,0] = u0
    vtx[0,:,0] = Dc0
    
    #Trace element initial concentrations
    utx[0,:,1] = 6. #Hf  %FCT from Bachmann etal JPet 2002.
    utx[0,:,2] = 21.9 # Y
    utx[0,:,3] = 3.9 # U
    utx[0,:,4] = 12.4 # Th
    utx[0,:,5] = 4. #Sm;%P
    utx[0,:,6] = 4. # Dy
    utx[0,:,7] = 4. # Yb
    vtx[0,:,1] = 0. #Hf 
    vtx[0,:,2] = 0. # Y
    vtx[0,:,3] = 0. # U
    vtx[0,:,4] = 0. # Th
    vtx[0,:,5] = 0. #Sm;%P
    vtx[0,:,6] = 0. # Dy
    vtx[0,:,7] = 0. # Yb

    
    Xstbl = np.zeros(nt)
    Xstbl[0] = Xsini*Rc*1e4
    Rtbl = np.ones(nt)*Rini
    CHstbl = np.zeros((nt,ntr))
    DCHstbl = np.zeros((nt,ntr))
    if affiche:
        fig, ax = plt.subplots(2,2)
    
    R = Rini
    Xs = Xsini
    Vj = 0
    Wj = 0
    jend = nt
    #### iterations in time
    for j in range(1,nt):
        ujm = utx[j-1,:,0]
        vjm = vtx[j-1,:,0]
        DRjm = DR[j-1]
        DXjm = DXs[j-1]
        V = Vj
        W = Wj
        u = utx[j-1,:,0]
        R0 = R
        Xs0 = Xs
        dXpdT = DXplag(Th[j])           
        dTdt = (Th[j]-Th[j-1])/dts
        gamma = {'Cm' : ZrSaturation(Th[j]), 'Dj' : DiffusionCoefficient(Th[j],xh2o)/Dscale,
                 'dXdt' : dXpdT*dTdt}
        for k in range(500): # iterations for the fixed point for the direct problem
            uold = u
            V, W, upper, lower, diagonal, sm = sc.functionF(gamma,ujm,u,R0, Xs0, xi, dts, constants)
            uu, nu2 = sc.thomas_vecteurs(lower,diagonal,upper,sm)
            u = np.array(uu)
            if np.linalg.norm(uold-u)**2<1e-20*np.array(nu2):
                break
        utx[j,:,0] = u
        DFV, DFW, DFAup, DFAlo, DFAdi, DFsm = sc.jacobianF(gamma,ujm,V,W,u,R0, Xs0, xi, dts, constants)
        
        Vj = V
        Wj = W
        Xs = np.max([Xs0+Vj*dts,1e-10])
        R = np.max([Xs+1e-15,R0+Wj*dts])
        if R-Xs<1e-12:
            print('No more melt - Xs=R, time step:',j)
            Xstbl[j:] = Xs*Rc*1e4
            Rtbl[j:] = R
            jend = j+1
            break
        VV[j] = Vj
        WW[j] = Wj
        Xstbl[j] = Xs*Rc*1e4
        Rtbl[j] = R
        
        DCm = DZrSaturation(Th[j])*alpha[j]
        DDj = DDiffusionCoefficient(Th[j],xh2o)/Dscale*alpha[j]
        DdXpdT= D2Xplag(Th[j])*alpha[j]
        DdX = DdXpdT*dTdt + dXpdT*(alpha[j]-alpha[j-1])/dts
        smv = -(DFAdi['1Cm']*DCm+DFAdi['1Dj']*DDj+DFAdi['1dX']*DdX)*u - DFAdi['6R']*DRjm*u - DFAdi['6Xs']*DXjm*u
        smv[:-1] = smv[:-1]-(DFAup['1Cm'][:-1]*DCm+DFAup['1Dj'][:-1]*DDj+DFAup['1dX'][:-1]*DdX)*u[1:] - DFAup['6R'][:-1]*DRjm*u[1:] - DFAup['6Xs'][:-1]*DXjm*u[1:]
        smv[1:] = smv[1:]-(DFAlo['1Cm'][1:]*DCm+DFAlo['1Dj'][1:]*DDj+DFAlo['1dX'][1:]*DdX)*u[:-1] - DFAlo['6R'][1:]*DRjm*u[:-1] - DFAlo['6Xs'][1:]*DXjm*u[:-1]
        smv += DFsm['1Cm']*DCm + DFsm['1Dj']*DDj + DFsm['1dX']*DdX + DFsm['2ujm']*vjm + DFsm['6R']*DRjm + DFsm['6Xs']*DXjm
        

        vold = vjm
        DuAvu = DFAdi['5u']*u
        DuAvu[:-1] = DuAvu[:-1]+DFAup['5u'][:-1]*u[1:]
        DuAvu[1:] = DuAvu[1:]+DFAlo['5u'][1:]*u[:-1]
        for k in range(500): # iterations for the fixed point for the tangent problem    
            smk = smv + DFsm['5u']*vold[1] - vold[1]*DuAvu
            vv, nv2 = sc.thomas_vecteurs(lower,diagonal,upper,smk)
            v = np.array(vv)
            if np.linalg.norm(vold-v)**2<1e-24*np.array(nv2):
                break
            vold = v
        vtx[j,:,0] = v
 
        DVV[j] = DFV['1Cm']*DCm + DFV['1Dj']*DDj + DFV['5u']*v[1]
        DWW[j] = DFW['1dX']*DdX
        DR[j] = DRjm + dts*DWW[j]
        DXs[j] = DXjm + dts*DVV[j]
        
        ### trace elements
        uTrjk,wTrjk = sc.derivativeFTr(xi,Vj,Wj,DVV[j], DWW[j],utx[j-1,:,1:],vtx[j-1,:,1:],Xs,DXs[j],R,DR[j],dts,Th[j],alpha[j],constants)
        utx[j,:,1:] = uTrjk
        ktr = Kdd(Th[j])
        Dktr = DKdd(Th[j])*alpha[j]
        CHstbl[j,:] = ktr[:ntr]*uTrjk[0,:]
           
        vtx[j,:,1:] = wTrjk
        DCHstbl[j,:] = ktr[:ntr]*wTrjk[0,:] + Dktr[:ntr]*uTrjk[0,:]
            
        Mpl = Mpl - 4*np.pi*R**2*Wj*dts*constants['kpl']*utx[j,-1,0]
        
        if np.mod(j,10)==0 and affiche:
            r=(Xs+(R-Xs)*xi)*Rc
            V0=Rc*1e4/tscale*tyear
            Mass = 4*np.pi*(Xs*Rc)**3*Czirc/3 + 4*np.pi*np.trapz(r**2*utx[j,:,0],r)+Mpl*Rc**3
            ax[0,0].plot(r*1e4,utx[j,:,0]); 
            ax[0,0].set_xlabel(r'r, \mu m')
            ax[0,0].set_ylabel('C, ppm')
            ax[1,0].plot(j*dts*tscale/tyear,(Mass-Mas0)/Mas0,'+r')
            ax[1,0].set_xlabel('time, years')
            ax[1,0].set_ylabel(r'\delta M')
            ax[0,1].plot(j*dts*tscale/tyear,V*V0,'+r',j*dts*tscale/tyear,-W*V0,'ob')
            ax[0,1].set_xlabel('time, years')
            ax[0,1].set_ylabel('V,-W, \mum/year')
            ax[1,1].plot(j*dts*tscale/tyear,Xs*Rc*1e4,'+r')
            ax[1,1].set_xlabel('time, years')
            ax[1,1].set_ylabel('Zr rad, \mum')
            
    return utx, vtx, VV, DVV, WW, DWW, Xstbl[:jend], Rtbl, CHstbl[:jend], DXs[:jend]*Rc*1e4, DCHstbl[:jend]



