#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan  3 17:30:41 2023

    Zircon diffusion inverse problem
    Copyright (C) 2024 Jerome Fehrenbach

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

library for the Stefan problem 
direct and tangent solve

includes the trace elements
uses cython to improve speed

This version of the library is adapted to the experimental data with 5 elts
the order of the elements is the following: 
['Hf','Y','U','Th','P']
    
@author: fehrenbach
"""

import solver_cython as sc
import numpy as np
import matplotlib.pyplot as plt


def ZrSaturation(T): #defining Zr saturation conditions
    Mfactor = 0.0000048*T**2-0.0083626*T+4.8484463; # empirical relations from magma
    Csat = 490000/np.exp(10108/T-1.16*(Mfactor-1)-1.48) # Boehnkeetal2013ChemGeol351,324 assuming 490,000 ppm in Zircon
    return Csat

def DZrSaturation(T): #derivative of Zr saturation conditions
    Mfactor = 0.0000048*T**2-0.0083626*T+4.8484463 # empirical relations from magma
    DMfactor = 2*0.0000048*T-0.0083626
    DCsat = 490000*(10108/T**2 + 1.16*DMfactor)*np.exp(-10108/T+1.16*(Mfactor-1)+1.48)
    return DCsat

def TemperatureHistory(Td,T0,time,n=0,amp=0):
    reltime = time/time[-1]
    Th = np.maximum(T0,Td-(Td-T0)*reltime-amp*np.sin(n*2*np.pi*reltime))
    return Th

def D1TemperatureHistory(Td,T0,time,n=0,amp=0):
    # derivative of temperature history wrt Td
    reltime = time/time[-1]
    DTh = 1-reltime
    DTh[np.where(T0>Td-(Td-T0)*reltime-amp*np.sin(n*2*np.pi*reltime))] = 0
    return DTh



def DiffusionCoefficient(T,x = 3):  
    #defining Zr diffusion coefficients in melts as f(T,XH2O)
    theta = 1000./T
    lnD = -(11.4*x+3.13)/(0.84*x+1)-(21.4*x+47)/(1.06*x+1)*theta#best fit of Zr Diff coefficients (several workers) and WH83 dependence on XH2O
    dif = np.exp(lnD)*1e4 # in cm2/s
    return dif

def DDiffusionCoefficient(T,x = 3):  
    #derivative of Zr diffusion coefficient in melts as f(T,XH2O)
    theta= 1000./T
    lnD=-(11.4*x+3.13)/(0.84*x+1)-(21.4*x+47)/(1.06*x+1)*theta#best fit of Zr Diff coefficients (several workers) and WH83 dependence on XH2O
    fact = (21.4*x+47)/(1.06*x+1)*1000/T**2
    Ddif = fact*np.exp(lnD)*1e4 # in cm2/s
    return Ddif

def TraceDiff_5elts(T, x=5):
    # Tr diffusion coefficient in melts as f(T,XH2O)
    # order of the elements: ['Hf','Y','U','Th','P']
    lnD_Hf = (-8.620340372*T-42705.17449-.318918919*x*T+4049.500765*x)/T
    lnD_Y = -9.925340370-(35587.24428-2615.214492*x)/T
    P = 0.02 #Pressure in GPa
    lnD_U = -6.37-2.65*np.sqrt(x) - (44729-1093*P-8944*np.sqrt(x))/T
    lnD_Th = -7.02-1.30*np.sqrt(x) - (44682-1370*P-7281*np.sqrt(x))/T
    lnD_P=((0.9469e1 * x - 0.1815e3) / (x + 0.2374e1) * ((1000 / T) - 0.54e0) - 0.267e2)
    Dif = 1e4*np.exp([lnD_Hf,lnD_Y,lnD_U,lnD_Th])
    return Dif

def DTraceDiff_5elts(T, x=5):
    # temperature derivative of Tr diffusion coefficient in melts as f(T,XH2O)
    lnD_Hf = (-8.620340372*T-42705.17449-.318918919*x*T+4049.500765*x)/T
    DHf = (-8.620340372-.318918919*x)/T-(-8.620340372*T-42705.17449-.318918919*x*T+4049.500765*x)/T**2
    lnD_Y = -9.925340370-(35587.24428-2615.214492*x)/T
    DY = (35587.24428-2615.214492*x)/T**2
    P = 0.02 #Pressure in GPa
    lnD_U = -6.37-2.65*np.sqrt(x) - (44729-1093*P-8944*np.sqrt(x))/T
    DU = (44729-1093*P-8944*np.sqrt(x))/T**2
    lnD_Th = -7.02-1.30*np.sqrt(x) - (44682-1370*P-7281*np.sqrt(x))/T
    DTh = (44682-1370*P-7281*np.sqrt(x))/T**2
    fact = np.array([DHf,DY,DU,DTh])
    DDif = fact*1e4*np.exp([lnD_Hf,lnD_Y,lnD_U,lnD_Th])
    return DDif



def Kdd_5elts(T):
    X=1000/T;
    KD_Hf = 11.29 * X - 2.275
    KD_Y = 19.47 * X - 13.04
    KD_U = 15.32 * X - 9.17
    KD_Th = 13.02 * X - 8.354
    kdd = np.exp(np.array([KD_Hf,KD_Y,KD_U,KD_Th]))
    return kdd

def DKdd_5elts(T):
    X=1000/T;
    KD_Hf = 11.29 * X - 2.275
    KD_Y = 19.47 * X - 13.04
    KD_U = 15.32 * X - 9.17
    KD_Th = 13.02 * X - 8.354
    fact = [-11.29*1000/T**2, -19.47*1000/T**2, -15.32*1000/T**2, -13.02*1000/T**2]
    Dkdd = fact*np.exp(np.array([KD_Hf,KD_Y,KD_U,KD_Th]))
    return Dkdd


def Xplag(T):
    x=T-273;
    Xpl = 1.94319815919224E-07*x**4+-6.44583169395759E-04*x**3+0.799105630147523*x**2+-439.159560503417*x+90385.6027471822
    Xpl=Xpl/100
    return Xpl

def DXplag(T):
    x=T-273;
    DXpl = 1.94319815919224E-07*4*x**3+-6.44583169395759E-04*3*x**2+0.799105630147523*x*2+-439.159560503417
    DXpl=DXpl/100
    return DXpl

def D2Xplag(T):
    x=T-273;
    D2Xpl = 1.94319815919224E-07*4*3*x**2+-6.44583169395759E-04*3*2*x+0.799105630147523*2
    D2Xpl=D2Xpl/100
    return D2Xpl



def functionF(gamma,ujm,u,R0, Xs0, xi, dt, constants):
    # calculation of the iteration, whose fixed point = sol. of the implicit time step
    #
    # IN : gamma = dict containing Cm, Dj, dXdt
    #       ujm = u_{j-1} Zr concentration at previous time step
    #       u : Zr concentration
    #       R0 / Xs0 : outer/inner radius
    #       xi : location of space points
    #       dt : time step
    #       constants : a dict containing the constant values
    #

    Cm = gamma['Cm']
    Dj = gamma['Dj']
    dXdt = gamma['dXdt']
    Czirc = constants['Czirc']
    kpl = constants['kpl'] 
    
    n = np.shape(xi)[0]
    xiplus = 0.5*xi
    xiplus[0:-1] += 0.5*xi[1:]
    xiplus[-1] = xi[-1]+0.5*(xi[-1]-xi[-2])
    ximoins = 0.5*xi
    ximoins[1:] += 0.5*xi[0:-1]
    ximoins[0] = -0.5*xi[1]

    Vk = -Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc)
    Wk = -dXdt/3/R0**2
 
    R = R0+Wk*dt
    Xs = Xs0+Vk*dt
    h = R-Xs
    
    upper = np.zeros(n)
    lower = np.zeros(n)
    diagonal = np.zeros(n)
    sm = np.zeros(n)
    
    diagonal[0] = 1
    sm[0] = Cm
    
    for i in range(1,n-1) :
        h12 = xi[i] - xi[i-1]
        h23 = xi[i+1] - xi[i]
        xi23 = xiplus[i]
        xi12 = ximoins[i]
        t10 = (2*Vk-Wk)*Xs - Vk*R
        upper[i] = -0.5*(1/h23 * (h23*h*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) - 2 * (xi23 * h + Xs) * Dj) * dt) 
        lower[i] =  0.5*(1 / h12 * dt * (h12 * h * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + 2 * (xi12 * h + Xs) * Dj))
        diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
        sm[i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))
  
    diagonal[n-1] = Dj + (xi[-1]-xi[-2])*(R-Xs)*Wk*(1-kpl)
    lower[n-1] = -Dj
    
    return Vk, Wk, upper, lower, diagonal, sm




def jacobianF(gamma,ujm,V,W,u,R0, Xs0, xi, dt, constants):
    # calculation of the jacobian of F
    #
    # IN : gamma = dict containing Cm, Dj, dXdt
    #       ujm = u_{j-1} Zr concentration at previous time step
    #       V, W, u : inner/outer velocity, Zr concentration
    #       R0 / Xs0 : outer/inner radius
    #       xi : location of space points
    #       dt : time step
    #       constants : a dict containing the constant values
    # OUT : DFV, DFW, DFAupper, DFAlower, DFAdiago, DFsm in dict structures with keys i
    #       where the keys i are :
    #           1Cm, 1Dj, 1dX, 2ujm, 3V, 4W, 5u, 6R, 6Xs
    #       the components that are zero are not computed
    #
    n = np.shape(xi)[0]
    DFV = {}
    DFW = {}
    DFAupper = {'1Cm':np.zeros(n), '1Dj':np.zeros(n),'1dX':np.zeros(n),'5u':np.zeros(n),'6R':np.zeros(n),'6Xs':np.zeros(n)}
    DFAlower = {'1Cm':np.zeros(n), '1Dj':np.zeros(n),'1dX':np.zeros(n),'5u':np.zeros(n),'6R':np.zeros(n),'6Xs':np.zeros(n)}
    DFAdiago = {'1Cm':np.zeros(n), '1Dj':np.zeros(n),'1dX':np.zeros(n),'5u':np.zeros(n),'6R':np.zeros(n),'6Xs':np.zeros(n)}
    DFsm = {'1Cm':np.zeros(n), '1Dj':np.zeros(n),'1dX':np.zeros(n),'2ujm':np.zeros(n),'5u':np.zeros(n),'6R':np.zeros(n),'6Xs':np.zeros(n)}

    Cm = gamma['Cm']
    Dj = gamma['Dj']
    dXdt = gamma['dXdt']
    Czirc = constants['Czirc']
    kpl = constants['kpl'] 
    
    xiplus = 0.5*xi
    xiplus[0:-1] += 0.5*xi[1:]
    xiplus[-1] = xi[-1]+0.5*(xi[-1]-xi[-2])
    ximoins = 0.5*xi
    ximoins[1:] += 0.5*xi[0:-1]
    ximoins[0] = -0.5*xi[1]

    Vk = -Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc)
    DFV['1Cm'] = Dj/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc) + Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc)**2
    DFV['1Dj'] = -(u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc)
    DFV['5u'] = -Dj /(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc) # note : derivative wrt u[1]
    DFV['6R'] = Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)**2/(Cm-Czirc)
    DFV['6Xs'] = -Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)**2/(Cm-Czirc)
    
    Wk = -dXdt/3/R0**2
    DFW['1dX'] = -1/3/R0**2
    DFW['6R'] = 2*dXdt/3/R0**3
    
    R = R0+Wk*dt
    Xs = Xs0+Vk*dt
    D6RR = 1+dt*DFW['6R']
    D6RXs = dt*DFV['6R']
    D6XsXs = 1+dt*DFV['6Xs']
    h = R-Xs
    D1Cmh = -dt*DFV['1Cm']
    D1Djh = -dt*DFV['1Dj']
    D1dXh = dt*DFW['1dX']
    D5uh = -dt*DFV['5u']
    D6Rh = D6RR-D6RXs
    D6Xsh = -D6XsXs
    
    upper = np.zeros(n)
    lower = np.zeros(n)
    diagonal = np.zeros(n)
    sm = np.zeros(n)
    
    diagonal[0] = 1
    sm[0] = Cm
    DFsm['1Cm'][0] = 1
    t10 = (2*Vk-Wk)*Xs - Vk*R
    D1Cmt10 = 2*DFV['1Cm']*Xs+(2*Vk-Wk)*dt*DFV['1Cm']-DFV['1Cm']*R
    D1Djt10 = 2*DFV['1Dj']*Xs+(2*Vk-Wk)*dt*DFV['1Dj']-DFV['1Dj']*R
    D1dXt10 = -DFW['1dX']*Xs - Vk*dt*DFW['1dX']
    D5ut10 = 2*DFV['5u']*Xs + (2*Vk-Wk)*dt*DFV['5u'] - DFV['5u']*R
    D6Rt10 = 2*DFV['6R']*Xs - DFV['6R']*R - Vk
    D6Xst10 = (2*Vk-Wk) + 2*DFV['6Xs']*Xs - DFV['6Xs']*R
    
    for i in range(1,n-1) :
        h12 = xi[i] - xi[i-1]
        h23 = xi[i+1] - xi[i]
        xi23 = xiplus[i]
        xi12 = ximoins[i]
        upper[i] = -0.5*dt/h23 * (h23*h*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) - 2 * (xi23 * h + Xs) * Dj)  
        DFAupper['1Cm'][i] = -0.5*dt/h23 * (h23*D1Cmh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D1Cmt10 + xi23**2 * (D1Cmh*(Vk-Wk) + h*DFV['1Cm'])  - DFV['1Cm']*Xs- Vk*dt*DFV['1Cm']) - 2 * (xi23 * D1Cmh + dt*DFV['1Cm']) * Dj)  
        DFAupper['1Dj'][i] = -0.5*dt/h23 * (h23*D1Djh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D1Djt10 + xi23**2 * (D1Djh*(Vk-Wk) + h*DFV['1Dj'])  - 1 - DFV['1Dj']*Xs- Vk*dt*DFV['1Dj'])  - 2 * (xi23 * D1Djh + dt*DFV['1Dj']) * Dj- 2 * (xi23 * h + Xs)) 
        DFAupper['1dX'][i] = -0.5*dt/h23 * (h23*D1dXh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D1dXt10 + xi23**2 * (D1dXh*(Vk-Wk) - h*DFW['1dX']))   - 2 * xi23 * D1dXh * Dj) 
        DFAupper['5u'][i]  = -0.5*dt/h23 * (h23*D5uh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D5ut10 + xi23**2 * (D5uh*(Vk-Wk)+ h*DFV['5u'])  - DFV['5u']*Xs - Vk*dt*DFV['5u']) - 2 * (xi23 * D5uh + dt*DFV['5u']) * Dj)  
        DFAupper['6R'][i]  = -0.5*dt/h23 * (h23*D6Rh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D6Rt10 + xi23**2 * D6Rh*(Vk-Wk)+ xi23**2 * h*(DFV['6R']-DFW['6R']) - DFV['6R']*Xs-Vk*D6RXs) - 2 * (xi23 * D6Rh+D6RXs) * Dj)
        DFAupper['6Xs'][i]  = -0.5*dt/h23 * (h23*D6Xsh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D6Xst10 + xi23**2 * D6Xsh*(Vk-Wk) + xi23**2 * h*DFV['6Xs'] - DFV['6Xs']*Xs- Vk*D6XsXs) - 2 * (xi23 * D6Xsh + D6XsXs) * Dj)
        
        lower[i] =  0.5*dt / h12 * (h12 * h * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + 2 * (xi12 * h + Xs) * Dj)
        DFAlower['1Cm'][i] = 0.5*dt / h12 * (h12 * D1Cmh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h12 * h * (xi12 * D1Cmt10 + xi12**2 * D1Cmh*(Vk-Wk) + xi12**2 * h*DFV['1Cm'] - DFV['1Cm']*Xs - Vk*dt*DFV['1Cm']) + 2 * (xi12 * D1Cmh + dt*DFV['1Cm']) * Dj)
        DFAlower['1Dj'][i] = 0.5*dt / h12 * (h12 * D1Djh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h12 * h * (xi12 * D1Djt10 + xi12**2 * D1Djh*(Vk-Wk) + xi12**2 * h*DFV['1Dj']  - 1 - DFV['1Dj']*Xs - Vk*dt*DFV['1Dj']) + 2 * (xi12 * D1Djh + dt*DFV['1Dj']) * Dj + 2 * (xi12 * h + Xs))
        DFAlower['1dX'][i] = 0.5*dt / h12 * (h12 * D1dXh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h12 * h * (xi12 * D1dXt10 + xi12**2 * D1dXh*(Vk-Wk) - xi12**2 * h*DFW['1dX'] )    + 2 * xi12 * D1dXh * Dj)
        DFAlower['5u'][i] = 0.5*dt / h12 * (h12 * D5uh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs)  + h12 * h * (xi12 * D5ut10 + xi12**2 * D5uh*(Vk-Wk)  + xi12**2 * h*DFV['5u']  - DFV['5u']*Xs - Vk*dt*DFV['5u'])   + 2 * (xi12 * D5uh + dt*DFV['5u']) * Dj)
        DFAlower['6R'][i] = 0.5*dt / h12 * (h12 * D6Rh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs)  + h12 * h * (xi12 * D6Rt10 + xi12**2 * D6Rh*(Vk-Wk) + xi12**2 * h*(DFV['6R']-DFW['6R']) - DFV['6R']*Xs - Vk*D6RXs) + 2 * (xi12 * D6Rh+D6RXs) * Dj)
        DFAlower['6Xs'][i] = 0.5*dt / h12 * (h12 * D6Xsh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h12 * h * (xi12 * D6Xst10 + xi12**2 * D6Xsh*(Vk-Wk) + xi12**2 * h*DFV['6Xs']  - DFV['6Xs']*Xs- Vk*D6XsXs) + 2 * (xi12 * D6Xsh + D6XsXs) * Dj)
        
        diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
        DFAdiago['1Cm'][i] = 0.5*(((-xi12 + 2 - xi23) * dt*DFV['1Cm'] ) * h**2 * (xi12 - xi23))+ 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1Cmh * (xi12 - xi23)) - DFAupper['1Cm'][i] - DFAlower['1Cm'][i]
        DFAdiago['1Dj'][i] = 0.5*(((-xi12 + 2 - xi23) * dt*DFV['1Dj']) * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1Djh * (xi12 - xi23)) - DFAupper['1Dj'][i] - DFAlower['1Dj'][i] 
        DFAdiago['1dX'][i] = 0.5*(( dt*DFW['1dX'] * (xi12 + xi23)) * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1dXh * (xi12 - xi23))- DFAupper['1dX'][i] - DFAlower['1dX'][i]
        DFAdiago['5u'][i] = 0.5*(((-xi12 + 2 - xi23) * dt*DFV['5u'] ) * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D5uh * (xi12 - xi23)) - DFAupper['5u'][i] - DFAlower['5u'][i]
        DFAdiago['6R'][i] = 0.5*(( (-xi12 + 2 - xi23) * D6RXs +(xi12 + xi23)) * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*D6Rh*h * (xi12 - xi23)) - DFAupper['6R'][i] - DFAlower['6R'][i]
        DFAdiago['6Xs'][i] = 0.5*((-xi12 + 2 - xi23) * D6XsXs  * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D6Xsh * (xi12 - xi23)) - DFAupper['6Xs'][i] - DFAlower['6Xs'][i]
        
        sm[i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))
        DFsm['1Cm'][i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * dt*DFV['1Cm'] ) * h**2 * (xi12 - xi23) + ((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1Cmh * (xi12 - xi23))
        DFsm['1Dj'][i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * dt*DFV['1Dj'] ) * h**2 * (xi12 - xi23) + ((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1Djh * (xi12 - xi23))
        DFsm['1dX'][i] = ujm[i]*0.5*((dt*DFW['1dX'] * (xi12 + xi23)) * h**2 * (xi12 - xi23) + ((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2 * h*D1dXh * (xi12 - xi23))
        DFsm['2ujm'][i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))
        DFsm['5u'][i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * dt*DFV['5u'] ) * h**2 * (xi12 - xi23) + ((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D5uh * (xi12 - xi23))
        DFsm['6R'][i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * D6RXs+D6RR*(xi12 + xi23)) * h**2 * (xi12 - xi23)) + ujm[i]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*D6Rh*h * (xi12 - xi23))
        DFsm['6Xs'][i] = ujm[i]*0.5*((-xi12 + 2 - xi23) * D6XsXs * h**2 * (xi12 - xi23)) + ujm[i]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * D6Xsh*2*h * (xi12 - xi23))
               
    diagonal[n-1] = Dj + (xi[-1]-xi[-2])*(R-Xs)*Wk*(1-kpl)
    DFAdiago['1Cm'][n-1] =  - dt*DFV['1Cm']*(xi[-1]-xi[-2])*Wk*(1-kpl)
    DFAdiago['1Dj'][n-1] =  1 - dt*DFV['1Dj']*(xi[-1]-xi[-2])*Wk*(1-kpl)
    DFAdiago['1dX'][n-1] = (xi[-1]-xi[-2])*dt*DFW['1dX']*Wk*(1-kpl)  + (xi[-1]-xi[-2])*(R-Xs)*DFW['1dX']*(1-kpl)
    DFAdiago['5u'][n-1] =  -dt*DFV['5u']*(xi[-1]-xi[-2])*Wk*(1-kpl)
    DFAdiago['6R'][n-1] = (xi[-1]-xi[-2])*Wk*(1-kpl) + (xi[-1]-xi[-2])*(R-Xs)*DFW['6R']*(1-kpl)
    DFAdiago['6Xs'][n-1] = -(xi[-1]-xi[-2])*D6XsXs*Wk*(1-kpl)
    lower[n-1] = -Dj
    DFAlower['1Dj'][n-1] = -1
    
    return DFV, DFW, DFAupper, DFAlower, DFAdiago, DFsm



def compute_G_DG(nt,nx,xi,dts,Xsini,Rini,Th,alpha,constants,Xsobs,CHstblobs,average):
    # compute G = the difference between the predictions and the obs.
    #       and DG.alpha the derivative of G in the direction alpha
    # 
    # IN : nt, nx : number of time steps/space steps
    #       xi : location of space points
    #       dts : value of the (scaled) time step
    #       Xsini : initial position of the crystal boundary
    #       Rini : initial outer radius of the domain
    #       Th : temperature history
    #       alpha : direction of variation of Th
    #       constants : a dict containing the constant values : 
    #           Rc - intial outer radius (length scale)
    #           Dscale - diffusion scale
    #           tscale - time scale
    #           tyear - duration of 1 year in seconds
    #           Czirc - concentration of ZR in crystal
    #           XH2O - content in H2O in %
    #       Xsobs : value of Xs for the observations
    #       uobs : value of uobs
    # OUT : Jobs, G, DG.alpha
    #       Jobs : locations of valid obs points
    #       G, DGalpha : value of G and derivative in the direction alpha
    #
    # The observations are the values of the concentrations of trace elts
    #       at the boundary of the crystal (Xs[:],CHstbl[:]) for all t
    # G and DG are obtained by interpolating these observations on the
    #       observations points Xsobs. For this it is necessary to set to 0
    #       the values outside the range of (Xs[:])
    #       The set of valid indices for Xsobs is denoted Jobs
    #       The size of G is ntr x nt : ntr=number of trace elts 
    #       
    ntr = 4 # number of trace elements
    
    utx, vtx, VV, DVV, WW, DWW, Xstbl, Rtbl, CHstbl, DXs, DCHstbl = solve_direct_tangent(nt,nx,xi,dts,Xsini,Rini,Th,alpha,constants,average)
    Jobs = np.where(np.logical_and(Xsobs>np.min(Xstbl),Xsobs<np.max(Xstbl)))[0]
    #print('DXs:',DXs)
    #print(np.linalg.norm(utx,axis=(0,1)),np.linalg.norm(vtx,axis=(0,1)))
    G = np.zeros((nt,ntr))
    DGalpha = np.zeros((nt,ntr))
    tt = np.zeros((nt,1))
    Dtt = np.zeros((nt,1))
    nxx = np.shape(Xstbl)[0]
    for k,j in enumerate(Jobs):
        i = np.max(np.where(Xsobs[j]>=Xstbl))
        ip1 = np.min([i+1,nxx-1])
        t = (Xstbl[ip1]-Xsobs[j])/(Xstbl[ip1]-Xstbl[i])
        #print(t)
        if np.isinf(t):
            t = 0.5
        tt[j] = t
        G[j,:] = t*CHstbl[i,:] + (1-t)*CHstbl[ip1,:] - CHstblobs[j,:]
        dt = (DXs[ip1]*(Xstbl[ip1]-Xstbl[i])-(Xstbl[ip1]-Xsobs[j])*(DXs[ip1]-DXs[i]))/(Xstbl[ip1]-Xstbl[i])**2
        if np.isnan(dt):
            dt = 0
        Dtt[j] = dt
        DLCHobs = dt*CHstbl[i,:] -dt*CHstbl[ip1,:] 
        DGalpha[j,:] = t*DCHstbl[i,:] + (1-t)*DCHstbl[ip1,:] + DLCHobs
        #print(dt,CHstbl[i,:],CHstbl[ip1,:],'DLCHobs:',np.linalg.norm(DLCHobs))
        #print(np.linalg.norm(DGalpha))
    finalradius = Xstbl[-1]
    Dfinalradius = DXs[-1]
    return Jobs, G, DGalpha, finalradius, Dfinalradius
    


def solve_direct_tangent(nt,nx,xi,dts,Xsini,Rini,Th,alpha,constants,ini_cond,affiche=False):
    # solve the direct problem
    # using implicit time discretization
    #
    # The direct and tangent model are solved as fixed point pbs with explicit
    #   functions to be iterated
    #
    # IN : nt, nx : number of time steps/space steps
    #       xi : location of space points
    #       dts : value of the (scaled) time step
    #       Xsini : initial position of the crystal boundary
    #       Rini : initial outer radius of the domain
    #       Th : temperature history
    #       alpha : direction of variation of Th
    #       constants : a dict containing the constant values : 
    #           Rc - intial outer radius (length scale)
    #           Dscale - diffusion scale
    #           tscale - time scale
    #           tyear - duration of 1 year in seconds
    #           Czirc - concentration of ZR in crystal
    #           XH2O - content in H2O in %
    #       ini_cond : initial condition for trace elements 
    #                   order of the elements: ['Hf','Y','U','Th','P']
    # OUT : utx : value of u(t,x,i) = concentration of species i at (t,x)
    #               for i=0 : Zr, other species for i>=1
    #       VV : value of the velocity V along the time
    #       WW : value of the velocity W along the time
    #       Xstbl /Rtbl/CHtbl: value of Xs (crystal radius)/R (outer radius)/Trace elts 
    #                           concentration along time
    ntr = 4 # number of trace elements
    c0 = ZrSaturation(Th[0])
    Dc0 = DZrSaturation(Th[0])*alpha[0]
    Rc = constants['Rc']
    Dscale = constants['Dscale']
    tscale = constants['tscale']
    tyear = constants['tyear']
    Czirc = constants['Czirc']
    xh2o = constants['XH2O']
    
    utx = np.zeros((nt,nx,ntr+1))
    vtx = np.zeros((nt,nx,ntr+1))
    VV = np.zeros(nt)
    WW = np.zeros(nt)
    DVV = np.zeros(nt)
    DWW = np.zeros(nt) 
    DR = np.zeros(nt)
    DXs = np.zeros(nt)
    
    Mpl = 0
    Mas0 = 4*np.pi*(Rc*Xsini)**3*Czirc/3 + 4*np.pi*c0*(Rc**3-Xsini**3)/3

    ###" Zr initial condition for t = 0
    u0 = np.zeros(nx)
    u0[:] = c0
    utx[0,:,0] = u0
    vtx[0,:,0] = Dc0
    
    #Trace element initial concentrations
    utx[0,:,1] = ini_cond[0] #Hf  
    utx[0,:,2] = ini_cond[1] # Y
    utx[0,:,3] = ini_cond[2] # U
    utx[0,:,4] = ini_cond[3] # Th
#    utx[0,:,5] = ini_cond[4] #Sm;%P
    vtx[0,:,1] = 0. #Hf 
    vtx[0,:,2] = 0. # Y
    vtx[0,:,3] = 0. # U
    vtx[0,:,4] = 0. # Th
#    vtx[0,:,5] = 0. #Sm;%P

    
    Xstbl = np.zeros(nt)
    Xstbl[0] = Xsini*Rc*1e4
    Rtbl = np.ones(nt)*Rini
    CHstbl = np.zeros((nt,ntr))
    DCHstbl = np.zeros((nt,ntr))
    if affiche:
        fig, ax = plt.subplots(2,2)
    
    R = Rini
    Xs = Xsini
    Vj = 0
    Wj = 0
    jend = nt
    #### iterations in time
    for j in range(1,nt):
        ujm = utx[j-1,:,0]
        vjm = vtx[j-1,:,0]
        DRjm = DR[j-1]
        DXjm = DXs[j-1]
        V = Vj
        W = Wj
        u = utx[j-1,:,0]
        if j%1000==0:
            print('iteration ',j,'/',nt)
 
        R0 = R
        Xs0 = Xs
        dXpdT = DXplag(Th[j])#(Xplag(Th[j]+1e-5)-Xplag(Th[j]-1e-5))/2e-5            
        dTdt = (Th[j]-Th[j-1])/dts
        gamma = {'Cm' : ZrSaturation(Th[j]), 'Dj' : DiffusionCoefficient(Th[j],xh2o)/Dscale,
                 'dXdt' : dXpdT*dTdt}
        for k in range(500):
            uold = u
            V, W, upper, lower, diagonal, sm = sc.functionF(gamma,ujm,u,R0, Xs0, xi, dts, constants)
            uu, nu2 = sc.thomas_vecteurs(lower,diagonal,upper,sm)
            u = np.array(uu)
            if np.linalg.norm(uold-u)**2<1e-20*np.array(nu2):
                break
        utx[j,:,0] = u
        DFV, DFW, DFAup, DFAlo, DFAdi, DFsm = sc.jacobianF(gamma,ujm,V,W,u,R0, Xs0, xi, dts, constants)
        
        Vj = V
        Wj = W
        Xs = np.max([Xs0+Vj*dts,1e-10])
        R = np.max([Xs+1e-15,R0+Wj*dts])
        if R-Xs<1e-12:
            print('No more melt - Xs=R, time step:',j)
            Xstbl[j:] = Xs*Rc*1e4
            Rtbl[j:] = R
            jend = j+1
            break
        VV[j] = Vj
        WW[j] = Wj
        Xstbl[j] = Xs*Rc*1e4
        Rtbl[j] = R
        
        DCm = DZrSaturation(Th[j])*alpha[j]
        DDj = DDiffusionCoefficient(Th[j],xh2o)/Dscale*alpha[j]
        DdXpdT= D2Xplag(Th[j])*alpha[j]
        DdX = DdXpdT*dTdt + dXpdT*(alpha[j]-alpha[j-1])/dts
        smv = -(DFAdi['1Cm']*DCm+DFAdi['1Dj']*DDj+DFAdi['1dX']*DdX)*u - DFAdi['6R']*DRjm*u - DFAdi['6Xs']*DXjm*u
        smv[:-1] = smv[:-1]-(DFAup['1Cm'][:-1]*DCm+DFAup['1Dj'][:-1]*DDj+DFAup['1dX'][:-1]*DdX)*u[1:] - DFAup['6R'][:-1]*DRjm*u[1:] - DFAup['6Xs'][:-1]*DXjm*u[1:]
        smv[1:] = smv[1:]-(DFAlo['1Cm'][1:]*DCm+DFAlo['1Dj'][1:]*DDj+DFAlo['1dX'][1:]*DdX)*u[:-1] - DFAlo['6R'][1:]*DRjm*u[:-1] - DFAlo['6Xs'][1:]*DXjm*u[:-1]
        #print(j,np.sum(smv),DDj)
        #print(np.sum(DFAdi['1Cm']),DCm,np.sum(DFAdi['1Dj']),DDj,np.sum(DFAdi['1dX']),DdX,np.sum(DFAdi['6R']),DRjm,np.sum(DFAdi['6Xs']),DXjm)
        smv += DFsm['1Cm']*DCm + DFsm['1Dj']*DDj + DFsm['1dX']*DdX + DFsm['2ujm']*vjm + DFsm['6R']*DRjm + DFsm['6Xs']*DXjm
        

        vold = vjm
        DuAvu = DFAdi['5u']*u
        DuAvu[:-1] = DuAvu[:-1]+DFAup['5u'][:-1]*u[1:]
        DuAvu[1:] = DuAvu[1:]+DFAlo['5u'][1:]*u[:-1]
        #print(np.linalg.norm(vold[1]),np.sum(diagonal),np.sum(lower),np.sum(upper),np.sum(smv))
        for k in range(500):     
            
            smk = smv + DFsm['5u']*vold[1] - vold[1]*DuAvu
            vv, nv2 = sc.thomas_vecteurs(lower,diagonal,upper,smk)
            v = np.array(vv)
            if np.linalg.norm(vold-v)**2<1e-24*np.array(nv2):
                break
            vold = v
        vtx[j,:,0] = v
        
        #print(DFV['1Cm'],DCm,DFV['1Dj'],DDj, DFV['5u'],v[1])
        if np.abs(v[1])>1e10:
            print(type(diagonal))
            plt.figure()
            plt.plot(smk,'r')
            plt.plot(smv,'r:')
            plt.plot(np.array(diagonal),'y')
            plt.plot(np.array(upper),'m:')
            plt.plot(np.array(lower),'g:')
            plt.plot(DFsm['5u'],'k')
            plt.plot(DuAvu,'c')
            print('----------------------------------')
        DVV[j] = DFV['1Cm']*DCm + DFV['1Dj']*DDj + DFV['5u']*v[1]
        DWW[j] = DFW['1dX']*DdX
        DR[j] = DRjm + dts*DWW[j]
        DXs[j] = DXjm + dts*DVV[j]
        
        ### trace elements
        #print('aaaaa',Xs,R,Xs<R)
        uTrjk = functionFTr(xi,Vj,Wj,utx[j-1,:,1:],Xs,R,dts,Th[j],constants)
        utx[j,:,1:] = uTrjk
        ktr = Kdd_5elts(Th[j])
        Dktr = DKdd_5elts(Th[j])*alpha[j]
        CHstbl[j,:] = ktr[:ntr]*uTrjk[0,:]
        
        wTrjk = derivativeFTr(xi,Vj,Wj,DVV[j], DWW[j],utx[j-1,:,1:],vtx[j-1,:,1:],Xs,DXs[j],R,DR[j],dts,Th[j],alpha[j],constants)
        vtx[j,:,1:] = wTrjk
        DCHstbl[j,:] = ktr[:ntr]*wTrjk[0,:] + Dktr[:ntr]*uTrjk[0,:]
            
        Mpl = Mpl - 4*np.pi*R**2*Wj*dts*constants['kpl']*utx[j,-1,0]
        
        if np.mod(j,10)==0 and affiche:
            r=(Xs+(R-Xs)*xi)*Rc
            V0=Rc*1e4/tscale*tyear
            Mass = 4*np.pi*(Xs*Rc)**3*Czirc/3 + 4*np.pi*np.trapz(r**2*utx[j,:,0],r)+Mpl*Rc**3
            ax[0,0].plot(r*1e4,utx[j,:,0]); 
            ax[0,0].set_xlabel(r'r, \mu m')
            ax[0,0].set_ylabel('C, ppm')
            ax[1,0].plot(j*dts*tscale/tyear,(Mass-Mas0)/Mas0,'+r')
            ax[1,0].set_xlabel('time, years')
            ax[1,0].set_ylabel(r'\delta M')
            ax[0,1].plot(j*dts*tscale/tyear,V*V0,'+r',j*dts*tscale/tyear,-W*V0,'ob')
            ax[0,1].set_xlabel('time, years')
            ax[0,1].set_ylabel('V,-W, \mum/year')
            ax[1,1].plot(j*dts*tscale/tyear,Xs*Rc*1e4,'+r')
            ax[1,1].set_xlabel('time, years')
            ax[1,1].set_ylabel('Zr rad, \mum')
            
    return utx, vtx, VV, DVV, WW, DWW, Xstbl[:jend], Rtbl, CHstbl[:jend], DXs[:jend]*Rc*1e4, DCHstbl[:jend]



def testF_DF(nt,nx,xi,dts,Xsini,Rini,Th,constants):
    # test of F and jacobianF
    #
    # IN : nt, nx : number of time steps/space steps
    #       xi : location of space points
    #       dts : value of the (scaled) time step
    #       Xs : initial position of the crystal boundary
    #       R : initial outer radius of the domain
    #       Th : temperature history
    #       constants : a dict containing the constant values : 
    #           Rc - intial outer radius (length scale)
    #           Dscale - diffusion scale
    #           tscale - time scale
    #           tyear - duration of 1 year in seconds
    #           Czirc - concentration of ZR in crystal
    #           XH2O - content in H2O in %
    ntr = 5 # number of trace elements
    c0 = ZrSaturation(Th[0])
    Dscale = constants['Dscale']
    xh2o = constants['XH2O']   
    utx = np.zeros((nt,nx,ntr+1))
    Dutx = np.zeros((nt,nx,ntr+1))

    ###" Zr initial condition for t = 0

    u0 = np.zeros(nx)
    u0[:] = c0
    utx[0,:,0] = u0
    
    #Trace element initial concentrations
    utx[0,:,1] = 6. #Hf  %FCT from Bachmann etal JPet 2002.
    utx[0,:,2] = 21.9 # Y
    utx[0,:,3] = 3.9 # U
    utx[0,:,4] = 12.4 # Th
    utx[0,:,5] = 4. #Sm;%P
    Dutx[0,:,1] = 0. #Hf  %FCT from Bachmann etal JPet 2002.
    Dutx[0,:,2] = 0. # Y
    Dutx[0,:,3] = 0. # U
    Dutx[0,:,4] = 0. # Th
    Dutx[0,:,5] = 0. #Sm;%P
    
    Xstbl = np.zeros(nt)
    Xstbl[0] = Xsini
    #Rtbl = np.ones(nt)*Rini
    #fig, ax = plt.subplots(2,2)
    
    R = Rini
    Xs = Xsini
    Vj = 0
    Wj = 0
    #### iterations in time

    for j in range(1,nt):
        ujm = utx[j-1,:,0]
        V = Vj
        W = Wj
        u = utx[j-1,:,0]
        #dXpdT=(Xplag(Th[j]+1e-5)-Xplag(Th[j]-1e-5))/2e-5
        dXpdT = DXplag(Th[j]+1e-5)
        #DdXpdT=(DXplag(Th[j]+1e-5)-DXplag(Th[j]-1e-5))/2e-5
        
        dTdt=(Th[j]-Th[j-1])/dts
        R0 = R
        Xs0 = Xs
        dTdt=(Th[j]-Th[j-1])/dts
        gamma = {'Cm' : ZrSaturation(Th[j]), 'Dj' : DiffusionCoefficient(Th[j],xh2o)/Dscale,
                 'dXdt' : dXpdT*dTdt}
        for k in range(500):
            uold = u
            #R = R0 + W*dts
            #Xs = Xs0 + V*dts
            V, W, upper, lower, diagonal, sm = sc.functionF(gamma,ujm,u,R0, Xs0, xi, dts, constants)
            uu, nu2 = sc.thomas_vecteurs(lower,diagonal,upper,sm)
            u = np.array(uu)
            if np.linalg.norm(uold-u)**2<1e-18*nu2:
                break
        Vj = V
        Wj = W
        utx[j,:,0] = u
        
        if j==2*nt//3:
            #np.random.seed(5524)
            beta = np.random.randn(7+2*nx)
            #beta[-1] = 0
            #beta[3:] = 0
            V0, W0, upper0, lower0, diagonal0, sm0 = sc.functionF(gamma,ujm,u,R0, Xs0, xi, dts, constants)
            DFV, DFW, DFAup, DFAlo, DFAdi, DFsm = sc.jacobianF(gamma,ujm,V,W,u,R0, Xs0, xi, dts, constants)
            upper0 = np.array(upper0)
            lower0 = np.array(lower0)
            diagonal0 = np.array(diagonal0)
            sm0 = np.array(sm0)
            
            for epsilon in [1e-2,1e-4,1e-6,1e-8,1e-10,1e-12]:
                gammah = gamma.copy()
                gammah['Cm'] = gamma['Cm']+epsilon*beta[0]
                gammah['Dj'] = gamma['Dj']+epsilon*beta[1]
                gammah['dXdt'] = gamma['dXdt']+epsilon*beta[2]
                ujmh = ujm + epsilon*beta[3:3+nx]
                Vh = V+epsilon*beta[3+nx]
                Wh = W+epsilon*beta[4+nx]
                uh = u+epsilon*beta[5+nx:5+2*nx]
                R0h = R0+epsilon*beta[5+2*nx]
                Xs0h = Xs0+epsilon*beta[6+2*nx]
                Vh, Wh, upperh, lowerh, diagonalh, smh = sc.functionF(gammah,ujmh,uh,R0h, Xs0h, xi, dts, constants)
                upperh = np.array(upperh)
                lowerh = np.array(lowerh)
                diagonalh = np.array(diagonalh)
                smh = np.array(smh)
            
                # derivative of FV
                res1 = DFV['1Cm']*beta[0] + DFV['1Dj']*beta[1] + DFV['5u']*beta[6+nx]+ DFV['6R']*beta[5+2*nx]+ DFV['6Xs']*beta[6+2*nx]
                difdiv = (Vh-V0)/epsilon
                print(epsilon,' V -rel. dif. ',(res1-difdiv)/res1)
                ## derivative of FW
                res1 = DFW['1dX']*beta[2] + DFW['6R']*beta[5+2*nx]
                difdiv = (Wh-W0)/epsilon
                print(epsilon,' W -rel. dif. ',(res1-difdiv)/res1)
                # derivative of FAupper
                res3 = DFAup['1Cm']*beta[0] + DFAup['1Dj']*beta[1] + DFAup['1dX']*beta[2] + DFAup['5u']*beta[6+nx] + DFAup['6R']*beta[5+2*nx]+ DFAup['6Xs']*beta[6+2*nx]
                difdiv3 = (upperh-upper0)/epsilon
                print(epsilon,'Aup - rel. dif. ',np.linalg.norm(res3-difdiv3)/np.linalg.norm(res3))
                # derivative of FAlower
                res3 = DFAlo['1Cm']*beta[0] + DFAlo['1Dj']*beta[1] + DFAlo['1dX']*beta[2] + DFAlo['5u']*beta[6+nx] + DFAlo['6R']*beta[5+2*nx]+ DFAlo['6Xs']*beta[6+2*nx]
                difdiv3 = (lowerh-lower0)/epsilon
                print(epsilon,'Alo - rel. dif. ',np.linalg.norm(res3-difdiv3)/np.linalg.norm(res3))
                # derivative of FAdiagonal
                res3 = DFAdi['1Cm']*beta[0] + DFAdi['1Dj']*beta[1] + DFAdi['1dX']*beta[2] + DFAdi['5u']*beta[6+nx] + DFAdi['6R']*beta[5+2*nx]+ DFAdi['6Xs']*beta[6+2*nx]
                difdiv3 = (diagonalh-diagonal0)/epsilon
                print(epsilon,'Adiago - rel. dif. ',np.linalg.norm(res3-difdiv3)/np.linalg.norm(res3))
                # derivative of sm
                res3 = DFsm['1Cm']*beta[0] + DFsm['1Dj']*beta[1] + DFsm['1dX']*beta[2] + DFsm['2ujm']*beta[3:3+nx]+ DFsm['5u']*beta[6+nx] + DFsm['6R']*beta[5+2*nx]+ DFsm['6Xs']*beta[6+2*nx]
                difdiv3 = (smh-sm0)/epsilon
                print(epsilon,'sm - rel. dif. ',np.linalg.norm(res3-difdiv3)/np.linalg.norm(res3))
                #tmp variable
                #res4 = Dtmp['1Cm']*beta[0] + Dtmp['1Dj']*beta[1] + Dtmp['1dX']*beta[2] + Dtmp['5u']*beta[6+nx] 
                #difdiv = (tmph-tmp0)/epsilon
                #print(epsilon,' tmp -rel. dif. ',np.linalg.norm(res4-difdiv)/np.linalg.norm(res4))
        
        Xs = np.max([Xs0+Vj*dts,1e-10])
        R = np.max([Xs+1e-15,R0+Wj*dts])

    return 




def testFTr_derivativeFTr(nt,nx,xi,dts,Xsini,Rini,Th,constants,alpha):
    # test of FTr and jacobianFTr
    #
    # IN : nt, nx : number of time steps/space steps
    #       xi : location of space points
    #       dts : value of the (scaled) time step
    #       Xs : initial position of the crystal boundary
    #       R : initial outer radius of the domain
    #       Th : temperature history
    #       constants : a dict containing the constant values : 
    #           Rc - intial outer radius (length scale)
    #           Dscale - diffusion scale
    #           tscale - time scale
    #           tyear - duration of 1 year in seconds
    #           Czirc - concentration of ZR in crystal
    #           XH2O - content in H2O in %
    ntr = 5 # number of trace elements
    c0 = ZrSaturation(Th[0])
    Dscale = constants['Dscale']
    xh2o = constants['XH2O']  
    Rc = constants['Rc']
    utx = np.zeros((nt,nx,ntr+1))
    Dutx = np.zeros((nt,nx,ntr+1))

    ###" Zr initial condition for t = 0

    u0 = np.zeros(nx)
    u0[:] = c0
    utx[0,:,0] = u0
    
    #Trace element initial concentrations
    utx[0,:,1] = 6. #Hf  %FCT from Bachmann etal JPet 2002.
    utx[0,:,2] = 21.9 # Y
    utx[0,:,3] = 3.9 # U
    utx[0,:,4] = 12.4 # Th
    utx[0,:,5] = 4. #Sm;%P
    Dutx[0,:,1] = 0. #Hf  %FCT from Bachmann etal JPet 2002.
    Dutx[0,:,2] = 0. # Y
    Dutx[0,:,3] = 0. # U
    Dutx[0,:,4] = 0. # Th
    Dutx[0,:,5] = 0. #Sm;%P

    
    Xstbl = np.zeros(nt)
    Xstbl[0] = Xsini
    #Rtbl = np.ones(nt)*Rini
    #fig, ax = plt.subplots(2,2)
    
    R = Rini
    Xs = Xsini
    DR = np.zeros(nt)
    DXs = np.zeros(nt)
    Vj = 0
    Wj = 0
    VV = np.zeros(nt)
    WW = np.zeros(nt)
    DVV = np.zeros(nt)
    DWW = np.zeros(nt) 
    Rtbl = np.ones(nt)*Rini
    #### iterations in time

    for j in range(1,nt):
        ujm = utx[j-1,:,0]
        vjm = Dutx[j-1,:,0]
        DRjm = DR[j-1]
        DXjm = DXs[j-1]
        V = Vj
        W = Wj
        u = utx[j-1,:,0]
        if j%1000==0:
            print('iteration ',j,'/',nt)
 
        R0 = R
        Xs0 = Xs
        dXpdT = DXplag(Th[j])#(Xplag(Th[j]+1e-5)-Xplag(Th[j]-1e-5))/2e-5            
        dTdt = (Th[j]-Th[j-1])/dts
        gamma = {'Cm' : ZrSaturation(Th[j]), 'Dj' : DiffusionCoefficient(Th[j],xh2o)/Dscale,
                 'dXdt' : dXpdT*dTdt}
        for k in range(500):
            uold = u
            V, W, upper, lower, diagonal, sm = sc.functionF(gamma,ujm,u,R0, Xs0, xi, dts, constants)
            uu, tmp = sc.thomas_vecteurs(lower,diagonal,upper,sm)
            u = np.array(uu)
            if np.linalg.norm(uold-u)<1e-12*np.linalg.norm(u):
                #print('sortie',k)
                break
        utx[j,:,0] = u
        DFV, DFW, DFAup, DFAlo, DFAdi, DFsm = sc.jacobianF(gamma,ujm,V,W,u,R0, Xs0, xi, dts, constants)
        
        Vj = V
        Wj = W
        
        VV[j] = Vj
        WW[j] = Wj
        Xstbl[j] = Xs*Rc*1e4
        Rtbl[j] = R
        Xs = np.max([Xs0+Vj*dts,1e-10])
        R = np.max([Xs+1e-15,R0+Wj*dts])
        
        DCm = DZrSaturation(Th[j])*alpha[j]
        DDj = DDiffusionCoefficient(Th[j],xh2o)/Dscale*alpha[j]
        DdXpdT= D2Xplag(Th[j])*alpha[j]
        DdX = DdXpdT*dTdt + dXpdT*(alpha[j]-alpha[j-1])/dts
        smv = -(DFAdi['1Cm']*DCm+DFAdi['1Dj']*DDj+DFAdi['1dX']*DdX)*u - DFAdi['6R']*DRjm*u - DFAdi['6Xs']*DXjm*u
        smv[:-1] = smv[:-1]-(DFAup['1Cm'][:-1]*DCm+DFAup['1Dj'][:-1]*DDj+DFAup['1dX'][:-1]*DdX)*u[1:] - DFAup['6R'][:-1]*DRjm*u[1:] - DFAup['6Xs'][:-1]*DXjm*u[1:]
        smv[1:] = smv[1:]-(DFAlo['1Cm'][1:]*DCm+DFAlo['1Dj'][1:]*DDj+DFAlo['1dX'][1:]*DdX)*u[:-1] - DFAlo['6R'][1:]*DRjm*u[:-1] - DFAlo['6Xs'][1:]*DXjm*u[:-1]
        #smv0 = smv
        smv += DFsm['1Cm']*DCm + DFsm['1Dj']*DDj + DFsm['1dX']*DdX + DFsm['2ujm']*vjm + DFsm['6R']*DRjm + DFsm['6Xs']*DXjm
        

        vold = vjm
        DuAvu = DFAdi['5u']*u
        DuAvu[:-1] = DuAvu[:-1]+DFAup['5u'][:-1]*u[1:]
        DuAvu[1:] = DuAvu[1:]+DFAlo['5u'][1:]*u[:-1]
        for k in range(500):     
            smk = smv + DFsm['5u']*vold[1] - vold[1]*DuAvu
            vv, tmp = sc.thomas_vecteurs(lower,diagonal,upper,smk)
            v = np.array(vv)
            if np.linalg.norm(vold-v)<1e-12*np.linalg.norm(v):
                #print('sortie',k)
                break
            if k%50==0:
                pass
                #print('iteration ',k,'/',np.linalg.norm(smk))
            vold = v
        Dutx[j,:,0] = v
 
        DVV[j] = DFV['1Cm']*DCm + DFV['1Dj']*DDj + DFV['5u']*v[1]
        DWW[j] = DFW['1dX']*DdX
        DR[j] = DRjm + dts*DWW[j]
        DXs[j] = DXjm + dts*DVV[j]
        
        ### trace elements
        uTrjk = functionFTr(xi,Vj,Wj,utx[j-1,:,1:],Xs,R,dts,Th[j],constants)
        utx[j,:,1:] = uTrjk
        wTrjk = derivativeFTr(xi,Vj,Wj,DVV[j], DWW[j],utx[j-1,:,1:],Dutx[j-1,:,1:],Xs,DXs[j],R,DR[j],dts,Th[j],alpha[j],constants)
        Dutx[j,:,1:] = wTrjk
        #print('rel. difference',np.linalg.norm(uTrjk-uuTrjk)/np.linalg.norm(uTrjk))
        #print('rel. difference',np.linalg.norm(lower-lower2)/np.linalg.norm(lower))
        
        if j==2*nt//3:
            #plt.figure()
            #plt.plot(lower,'ko')
            #plt.plot(lower2,'r+')
            #np.random.seed(5524)
            beta = np.random.randn(7+2*nx)
            #beta[-1] = 0
            #beta[3:] = 0
            uTrjk = functionFTr(xi,Vj,Wj,utx[j-1,:,1:],Xs,R,dts,Th[j],constants)
            wTrjk = derivativeFTr(xi,Vj,Wj,DVV[j], DWW[j],utx[j-1,:,1:],Dutx[j-1,:,1:],Xs,DXs[j],R,DR[j],dts,Th[j],alpha[j],constants)
            
            
            for epsilon in [1e-2,1e-4,1e-6,1e-8,1e-10,1e-12]:
                gammah = gamma.copy()
                gammah['Cm'] = gamma['Cm']+epsilon*beta[0]
                gammah['Dj'] = gamma['Dj']+epsilon*beta[1]
                gammah['dXdt'] = gamma['dXdt']+epsilon*beta[2]
                ujmh = ujm + epsilon*beta[3:3+nx]
                Vh = V+epsilon*beta[3+nx]
                Wh = W+epsilon*beta[4+nx]
                uh = u+epsilon*beta[5+nx:5+2*nx]
                R0h = R0+epsilon*beta[5+2*nx]
                Xs0h = Xs0+epsilon*beta[6+2*nx]
                Vh, Wh, upperh, lowerh, diagonalh, smh = sc.functionF(gammah,ujmh,uh,R0h, Xs0h, xi, dts, constants)
                uTrjkh = functionFTr(xi,Vh,Wh,utx[j-1,:,1:],Xs,R,dts,Th[j],constants)
                # derivative of FV
                res1 = DFV['1Cm']*beta[0] + DFV['1Dj']*beta[1] + DFV['5u']*beta[6+nx]+ DFV['6R']*beta[5+2*nx]+ DFV['6Xs']*beta[6+2*nx]
                difdiv = (Vh-Vj)/epsilon
                #print(epsilon,' V -rel. dif. ',(res1-difdiv)/res1)
                ## derivative of FW
                res1 = DFW['1dX']*beta[2] + DFW['6R']*beta[5+2*nx]
                difdiv = (Wh-Wj)/epsilon
                print(epsilon,' W -rel. dif. ',(res1-difdiv)/res1)
                # derivative of FAupper
                res3 = DFAup['1Cm']*beta[0] + DFAup['1Dj']*beta[1] + DFAup['1dX']*beta[2] + DFAup['5u']*beta[6+nx] + DFAup['6R']*beta[5+2*nx]+ DFAup['6Xs']*beta[6+2*nx]
                difdiv3 = (np.array(upperh)-np.array(upper))/epsilon
                print(epsilon,'Aup - rel. dif. ',np.linalg.norm(res3-difdiv3)/np.linalg.norm(res3))
                # derivative of FAlower
                res3 = DFAlo['1Cm']*beta[0] + DFAlo['1Dj']*beta[1] + DFAlo['1dX']*beta[2] + DFAlo['5u']*beta[6+nx] + DFAlo['6R']*beta[5+2*nx]+ DFAlo['6Xs']*beta[6+2*nx]
                difdiv3 = (np.array(lowerh)-np.array(lower))/epsilon
                print(epsilon,'Alo - rel. dif. ',np.linalg.norm(res3-difdiv3)/np.linalg.norm(res3))
                # derivative of FAdiagonal
                res3 = DFAdi['1Cm']*beta[0] + DFAdi['1Dj']*beta[1] + DFAdi['1dX']*beta[2] + DFAdi['5u']*beta[6+nx] + DFAdi['6R']*beta[5+2*nx]+ DFAdi['6Xs']*beta[6+2*nx]
                difdiv3 = (np.array(diagonalh)-np.array(diagonal))/epsilon
                print(epsilon,'Adiago - rel. dif. ',np.linalg.norm(res3-difdiv3)/np.linalg.norm(res3))
                # derivative of sm
                res3 = DFsm['1Cm']*beta[0] + DFsm['1Dj']*beta[1] + DFsm['1dX']*beta[2] + DFsm['2ujm']*beta[3:3+nx]+ DFsm['5u']*beta[6+nx] + DFsm['6R']*beta[5+2*nx]+ DFsm['6Xs']*beta[6+2*nx]
                difdiv3 = (np.array(smh)-np.array(sm))/epsilon
                print(epsilon,'sm - rel. dif. ',np.linalg.norm(res3-difdiv3)/np.linalg.norm(res3))
                #tmp variable
                #res4 = Dtmp['1Cm']*beta[0] + Dtmp['1Dj']*beta[1] + Dtmp['1dX']*beta[2] + Dtmp['5u']*beta[6+nx] 
                #difdiv = (tmph-tmp0)/epsilon
                #print(epsilon,' tmp -rel. dif. ',np.linalg.norm(res4-difdiv)/np.linalg.norm(res4))
        
        Xs = np.max([Xs0+Vj*dts,1e-10])
        R = np.max([Xs+1e-15,R0+Wj*dts])

    return 




def functionFTr(xi,Vj,Wj,ujm,Xs,R,dts,Thj,constants):
    # one iteration in time of the FV scheme, for Tr elements
    # IN : xi : spatials discretization points
    #       Vj/Wj - velocity of the inner/outer boundary
    #       ujm : uTr at the time step j-1
    #       Xs : position of the interface crystal/melt 
    #       R : outer radius of the domain
    #       dts : time step
    #       Thj : temperature at time tj
    #       constants : a dict containing the constant values
    # OUT :    
    #       uTr_j : uTr at the next time step
    #
    # the space variable is  x\in (0,1)
 
    xh2o = constants['XH2O']
    n = np.shape(xi)[0]
    ntr = np.shape(ujm)[1]
    uTrjk = np.zeros((n,ntr))
    
    DifTr = TraceDiff_5elts(Thj,xh2o)/constants['Dscale']
    kt = Kdd_5elts(Thj)
    xiplus = 0.5*xi
    xiplus[0:-1] += 0.5*xi[1:]
    xiplus[-1] = xi[-1]+0.5*(xi[-1]-xi[-2])
    ximoins = 0.5*xi
    ximoins[1:] += 0.5*xi[0:-1]
    ximoins[0] = -0.5*xi[1]
    h = R-Xs
    kpl = 0
    for k in range(ntr):
        ktr = kt[k]
        Difk = DifTr[k]
        
        upper = np.zeros(n)
        lower = np.zeros(n)
        diagonal = np.zeros(n)
        sm = np.zeros(n)
        
        t10 = (2*Vj-Wj)*Xs - Vj*R
        for i in range(1,n-1) :
            h12 = xi[i] - xi[i-1]
            h23 = xi[i+1] - xi[i]
            xi23 = xiplus[i]
            xi12 = ximoins[i]
            
            upper[i] = -0.5*(1/h23 * (h23*h*(xi23 * t10 + xi23**2 * h*(Vj-Wj) - Difk - Vj*Xs) - 2 * (xi23 * h + Xs) * Difk) * dts) 
            lower[i] =  0.5*(1 / h12 * dts * (h12 * h * (xi12 * t10 + xi12**2 * h*(Vj-Wj) - Difk - Vj*Xs) + 2 * (xi12 * h + Xs) * Difk))
            diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
            sm[i] = ujm[i,k]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))

        diagonal[0] = Difk - (xi[1]-xi[0])*(R-Xs)*Vj*(1-ktr)
        upper[0] = -Difk
        
        diagonal[n-1] = Difk + (xi[-1]-xi[-2])*(R-Xs)*Wj*(1-kpl)
        lower[n-1] = -Difk
        uTrjk[:,k], tmp =  sc.thomas_vecteurs(lower,diagonal,upper,sm)
 
    return uTrjk



def derivativeFTr(xi,Vj,Wj,DVj, DWj,ujm,vjm,Xs,DXs,R,DR,dts,Thj,alphaj,constants):
    # one iteration in time of the FV scheme, for the derivative of Tr elements 
    
    # IN : xi : spatials discretization points
    #       Vj/Wj - velocity of the inner/outer boundary
    #       DVj/DWj - derivatives of the velocities
    #       ujm : uTr at the time step j-1
    #       vjm : derivative of ujm at time step j-1
    #       Xs : position of the interface crystal/melt
    #       DXs : derivative of Xs
    #       R : outer radius of the domain
    #       DR : derivative of R
    #       dts : time step
    #       Thj : temperature at time tj
    #       alphaj : variation of the temperature at time j
    #       constants : a dict containing the constant values
    # OUT :    
    #       wTr_j : the derivative uTr at the next time step
    #
    # the space variable is  x\in (0,1)
 
    xh2o = constants['XH2O']
    n = np.shape(xi)[0]
    ntr = np.shape(ujm)[1]
    uTrjk = np.zeros((n,ntr))
    wTrjk = np.zeros((n,ntr))

    DifTr = TraceDiff_5elts(Thj,xh2o)/constants['Dscale']
    DDifTr = DTraceDiff_5elts(Thj,xh2o)/constants['Dscale']*alphaj
    kt = Kdd_5elts(Thj)
    Dkt = DKdd_5elts(Thj)*alphaj
    xiplus = 0.5*xi
    xiplus[0:-1] += 0.5*xi[1:]
    xiplus[-1] = xi[-1]+0.5*(xi[-1]-xi[-2])
    ximoins = 0.5*xi
    ximoins[1:] += 0.5*xi[0:-1]
    ximoins[0] = -0.5*xi[1]
    h = R - Xs
    Dh = DR - DXs
    kpl = 0
    
    for k in range(ntr):
        ktr = kt[k]
        Difk = DifTr[k]
        Dktr = Dkt[k]
        DDifk = DDifTr[k]
        
        upper = np.zeros(n)
        lower = np.zeros(n)
        diagonal = np.zeros(n)
        sm = np.zeros(n)
        Dupper = np.zeros(n)
        Dlower = np.zeros(n)
        Ddiagonal = np.zeros(n)
        Dsm = np.zeros(n)
        
        t10 = (2*Vj-Wj)*Xs - Vj*R
        Dt10 = (2*DVj-DWj)*Xs + (2*Vj-Wj)*DXs - DVj*R - Vj*DR
        for i in range(1,n-1) :
            h12 = xi[i] - xi[i-1]
            h23 = xi[i+1] - xi[i]
            xi23 = xiplus[i]
            xi12 = ximoins[i]
            
            upper[i] = -0.5* dts/h23 * (h23*h*(xi23 * t10 + xi23**2 * h*(Vj-Wj) - Difk - Vj*Xs) - 2 * (xi23 * h + Xs) * Difk) 
            Dupper[i] = -0.5*dts/h23 * (h23*Dh*(xi23 * t10 + xi23**2 * h*(Vj-Wj) - Difk - Vj*Xs) + h23*h*(xi23 * Dt10 + xi23**2 * Dh*(Vj-Wj)+ xi23**2 * h*(DVj-DWj) - DDifk - DVj*Xs- Vj*DXs) - 2 * (xi23 * Dh + DXs) * Difk- 2 * (xi23 * h + Xs) * DDifk) 
            
            lower[i] =  0.5*(1 / h12 * dts * (h12 * h * (xi12 * t10 + xi12**2 * h*(Vj-Wj) - Difk - Vj*Xs) + 2 * (xi12 * h + Xs) * Difk))
            Dlower[i] =  0.5 / h12 * dts * (h12 * Dh * (xi12 * t10 + xi12**2 * h*(Vj-Wj) - Difk - Vj*Xs)+h12 * h * (xi12 * Dt10 + xi12**2 * Dh*(Vj-Wj)+ xi12**2 * h*(DVj-DWj) - DDifk - DVj*Xs- Vj*DXs) + 2 * (xi12 * Dh + DXs) * Difk+ 2 * (xi12 * h + Xs) * DDifk)
            
            diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
            Ddiagonal[i] = 0.5*((-xi12 + 2 - xi23) * DXs + DR * (xi12 + xi23)) * h**2 * (xi12 - xi23)+0.5*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*Dh * (xi12 - xi23) - Dupper[i] - Dlower[i]
                     
            sm[i] = ujm[i,k]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))
            Dsm[i] = vjm[i,k]*0.5*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23) + ujm[i,k]*0.5*((-xi12 + 2 - xi23) * DXs + DR * (xi12 + xi23)) * h**2 * (xi12 - xi23) + ujm[i,k]*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h*Dh * (xi12 - xi23)


        diagonal[0] = Difk - (xi[1]-xi[0])*(R-Xs)*Vj*(1-ktr)
        Ddiagonal[0] = DDifk - (xi[1]-xi[0])*(DR-DXs)*Vj*(1-ktr) - (xi[1]-xi[0])*(R-Xs)*DVj*(1-ktr) + (xi[1]-xi[0])*(R-Xs)*Vj*Dktr
        
        upper[0] = -Difk
        Dupper[0] = -DDifk
        
        diagonal[n-1] = Difk + (xi[-1]-xi[-2])*(R-Xs)*Wj*(1-kpl)
        Ddiagonal[n-1] = DDifk + (xi[-1]-xi[-2])*(DR-DXs)*Wj*(1-kpl)+ (xi[-1]-xi[-2])*(R-Xs)*DWj*(1-kpl)
        
        lower[n-1] = -Difk
        Dlower[n-1] = -DDifk
        
       
        uTrjk[:,k], tmp =  sc.thomas_vecteurs(lower,diagonal,upper,sm)
        Aprime_u = Ddiagonal*uTrjk[:,k]
        Aprime_u[:-1] += Dupper[:-1]*uTrjk[1:,k]
        Aprime_u[1:] += Dlower[1:]*uTrjk[:-1,k]
        
        sm2 = -Aprime_u + Dsm
        wTrjk[:,k], tmp =  sc.thomas_vecteurs(lower,diagonal,upper,sm2)
    return wTrjk




def derivativeFTrtime(xi,Vj,Wj,DVj,uj,ujm,wjm,Xs,DXs,R,dts,Thj,constants):
    # one iteration in time of the FV scheme, for the derivative of Tr elements wrt a=log(Delta t)

    # IN : xi : spatials discretization points
    #       Vj/Wj - velocity of the inner/outer boundary
    #       DVj - derivatives of the velocity
    #       uj : uTr at the time step j
    #       ujm : uTr at the time step j-1
    #       wjm : derivative of uTr at time step j-1
    #       Xs : position of the interface crystal/melt
    #       DXs : derivative of Xs
    #       R : outer radius of the domain
    #       dts : time step
    #       Thj : temperature at time tj
    #       constants : a dict containing the constant values
    # OUT :    
    #       wTr_j : the derivative uTr at the next time step
    #
    # the space variable is  x\in (0,1)
        
    xh2o = constants['XH2O']
    n = np.shape(xi)[0]
    ntr = np.shape(uj)[1]
    wTrj = np.zeros((n,ntr))

    DifTr = TraceDiff_5elts(Thj,xh2o)/constants['Dscale']
    kt = Kdd_5elts(Thj)
    xiplus = 0.5*xi
    xiplus[0:-1] += 0.5*xi[1:]
    xiplus[-1] = xi[-1]+0.5*(xi[-1]-xi[-2])
    ximoins = 0.5*xi
    ximoins[1:] += 0.5*xi[0:-1]
    ximoins[0] = -0.5*xi[1]
    h = R - Xs
    Dh = - DXs
    t10 = (2*Vj-Wj)*Xs - Vj*R
    Dt10 = (2*DVj+Wj)*Xs + (2*Vj-Wj)*DXs - DVj*R
    kpl = 0
    
    for k in range(ntr):
        ktr = kt[k]
        Difk = DifTr[k]
        upper = np.zeros(n)
        lower = np.zeros(n)
        diagonal = np.zeros(n)
        Lsm = np.zeros(n)
         
        Dupper = np.zeros(n)
        Dlower = np.zeros(n)
        Ddiagonal = np.zeros(n)
        DLsm = np.zeros(n)

        for i in range(1,n-1):
            h12 = xi[i] - xi[i-1]
            h23 = xi[i+1] - xi[i]
            xi23 = xiplus[i]
            xi12 = ximoins[i]
            
            upper[i] = -0.5*dts/h23 * (h23*h*(xi23 * t10 + xi23**2 * h*(Vj-Wj) - Difk - Vj*Xs) - 2 * (xi23 * h + Xs) * Difk)
            Dupper[i] = upper[i] - 0.5*dts/h23 * (h23*Dh*(xi23 * t10 + xi23**2 * h*(Vj-Wj) - Difk - Vj*Xs)+h23*h*(xi23 * Dt10 + xi23**2 * Dh*(Vj-Wj)+ xi23**2 * h*(DVj+Wj) - DVj*Xs- Vj*DXs) - 2 * (xi23 * Dh + DXs) * Difk)
            
            lower[i] =  0.5/ h12 * dts * (h12 * h * (xi12 * t10 + xi12**2 * h*(Vj-Wj) - Difk - Vj*Xs) + 2 * (xi12 * h + Xs) * Difk)
            Dlower[i] =  lower[i] + 0.5/ h12 * dts * (h12 * Dh * (xi12 * t10 + xi12**2 * h*(Vj-Wj) - Difk - Vj*Xs)+h12 * h * (xi12 * Dt10 + xi12**2 * Dh*(Vj-Wj)+ xi12**2 * h*(DVj+Wj) - DVj*Xs- Vj*DXs) + 2 * (xi12 * Dh + DXs) * Difk)
                  
            diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
            Ddiagonal[i] = 0.5*((-xi12 + 2 - xi23) * DXs * h**2 * (xi12 - xi23))+0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*Dh * (xi12 - xi23)) - Dupper[i] - Dlower[i]
            
            Lsm[i] = 0.5*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)
            DLsm[i] = 0.5*(-xi12 + 2 - xi23) * DXs * h**2 * (xi12 - xi23)+0.5*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*Dh * (xi12 - xi23)

        diagonal[0] = Difk - (xi[1]-xi[0])*(R-Xs)*Vj*(1-ktr)   
        Ddiagonal[0] = (xi[1]-xi[0])*DXs*Vj*(1-ktr)- (xi[1]-xi[0])*(R-Xs)*DVj*(1-ktr)    
        
        upper[0] = -Difk        
        diagonal[n-1] = Difk + (xi[-1]-xi[-2])*(R-Xs)*Wj*(1-kpl)        
        lower[n-1] = -Difk
        Ddiagonal[n-1] =  -(xi[-1]-xi[-2])*DXs*Wj*(1-kpl) + (xi[-1]-xi[-2])*(R-Xs)*(-Wj)*(1-kpl)   
        
        
        smw = Lsm*wjm[:,k] + DLsm*ujm[:,k] 
        smw -= Ddiagonal*uj[:,k]
        smw[:-1] = smw[:-1] - Dupper[:-1]*uj[1:,k]
        smw[1:] = smw[1:] - Dlower[1:]*uj[:-1,k]
        wTrj[:,k], tmp =  sc.thomas_vecteurs(lower,diagonal,upper,smw)
      
    return wTrj#, upper, Dupper

