#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 21 11:40:54 2023

    Zircon diffusion inverse problem
    Copyright (C) 2024 Jerome Fehrenbach

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.


@author: fehren
"""

import numpy as np
cimport numpy as np
import lib_stefan_trace as ls

def thomas_vecteurs(double[:] lower,double[:] diago,double[:] upper,double[:] sm):
    # solution of a tridiagonal system using Thomas' method
    cdef int n= sm.shape[0]
    cdef int i
    cdef double nx2 = 0.
    cdef double[:] cp = np.zeros(n-1)
    cdef double[:] dp = np.zeros(n)
    cdef double[:] x = np.zeros(n)
    cp[0] = upper[0]/diago[0]
    dp[0] = sm[0]/diago[0]
    for i in range(1,n-1):
        cp[i] = upper[i]/(diago[i]-lower[i]*cp[i-1])
        dp[i] = (sm[i]-lower[i]*dp[i-1])/(diago[i]-lower[i]*cp[i-1])
    x[n-1] = (sm[n-1]-lower[n-1]*dp[n-2])/(diago[n-1]-lower[n-1]*cp[n-2])
    for i in range(n-2,-1,-1):
        x[i] = dp[i] - cp[i]*x[i+1]
    for i in range(n):
        nx2 += x[i]**2
    return x, nx2



def jacobianF(gamma,double[:] ujm,double V,double W,double[:] u,double R0,double  Xs0,double[:] xi,double  dt, constants):
    # calculation of the jacobian of F
    #
    # IN : gamma = dict containing Cm, Dj, dXdt
    #       ujm = u_{j-1} Zr concentration at previous time step
    #       V, W, u : inner/outer velocity, Zr concentration
    #       R / Xs : outer/inner radius
    #       xi : location of space points
    #       dt : time step
    #       constants : a dict containing the constant values
    # OUT : DFV, DFW, DFAupper, DFAlower, DFAdiago, DFsm in dict structures with keys i
    #       where the keys i are :
    #           1Cm, 1Dj, 1dX, 2ujm, 3V, 4W, 5u, 6R, 6Xs
    #       the components that are zero are not computed
    #
    cdef int n = np.shape(xi)[0]
    cdef int i
    cdef double h12, h23, xi12, xi23
    DFV = {}
    DFW = {}
    DFAupper = {'1Cm':np.zeros(n), '1Dj':np.zeros(n),'1dX':np.zeros(n),'5u':np.zeros(n),'6R':np.zeros(n),'6Xs':np.zeros(n)}
    DFAlower = {'1Cm':np.zeros(n), '1Dj':np.zeros(n),'1dX':np.zeros(n),'5u':np.zeros(n),'6R':np.zeros(n),'6Xs':np.zeros(n)}
    DFAdiago = {'1Cm':np.zeros(n), '1Dj':np.zeros(n),'1dX':np.zeros(n),'5u':np.zeros(n),'6R':np.zeros(n),'6Xs':np.zeros(n)}
    DFsm = {'1Cm':np.zeros(n), '1Dj':np.zeros(n),'1dX':np.zeros(n),'2ujm':np.zeros(n),'5u':np.zeros(n),'6R':np.zeros(n),'6Xs':np.zeros(n)}

    cdef double Cm = gamma['Cm']
    cdef double Dj = gamma['Dj']
    cdef double dXdt = gamma['dXdt']
    cdef double Czirc = constants['Czirc']
    cdef double kpl = constants['kpl'] 
    
    cdef double[:] xiplus=np.zeros(n)
    cdef double[:] ximoins=np.zeros(n)
    for i in range(n):
        xiplus[i] = 0.5*xi[i]
        ximoins[i] = 0.5*xi[i]
    for i in range(n-1):
        xiplus[i] += 0.5*xi[i+1]
        ximoins[i+1] += 0.5*xi[i]
    xiplus[n-1] = xi[n-1]+0.5*(xi[n-1]-xi[n-2])

    cdef double Vk = -Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc)
    DFV['1Cm'] = Dj/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc) + Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc)**2
    DFV['1Dj'] = -(u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc)
    DFV['5u'] = -Dj /(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc) # note : derivative wrt u[1]
    DFV['6R'] = Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)**2/(Cm-Czirc)
    DFV['6Xs'] = -Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)**2/(Cm-Czirc)
    
    cdef double Wk = -dXdt/3/R0**2
    DFW['1dX'] = -1./3/R0**2
    DFW['6R'] = 2*dXdt/3/R0**3
    
    cdef double R = R0+Wk*dt
    cdef double Xs = Xs0+Vk*dt
    cdef double D6RR = 1+dt*DFW['6R']
    cdef double D6RXs = dt*DFV['6R']
    cdef double D6XsXs = 1+dt*DFV['6Xs']
    cdef double h = R-Xs
    cdef double D1Cmh = -dt*DFV['1Cm']
    cdef double D1Djh = -dt*DFV['1Dj']
    cdef double D1dXh = dt*DFW['1dX']
    cdef double D5uh = -dt*DFV['5u']
    cdef double D6Rh = D6RR-D6RXs
    cdef double D6Xsh = -D6XsXs
    
    cdef double[:] upper = np.zeros(n)
    cdef double[:] lower = np.zeros(n)
    cdef double[:] diagonal = np.zeros(n)
    cdef double[:] sm = np.zeros(n)
    
    diagonal[0] = 1
    sm[0] = Cm
    DFsm['1Cm'][0] = 1
    cdef double t10 = (2*Vk-Wk)*Xs - Vk*R
    cdef double D1Cmt10 = 2*DFV['1Cm']*Xs+(2*Vk-Wk)*dt*DFV['1Cm']-DFV['1Cm']*R
    cdef double D1Djt10 = 2*DFV['1Dj']*Xs+(2*Vk-Wk)*dt*DFV['1Dj']-DFV['1Dj']*R
    cdef double D1dXt10 = -DFW['1dX']*Xs - Vk*dt*DFW['1dX']
    cdef double D5ut10 = 2*DFV['5u']*Xs + (2*Vk-Wk)*dt*DFV['5u'] - DFV['5u']*R
    cdef double D6Rt10 = 2*DFV['6R']*Xs - DFV['6R']*R - Vk
    cdef double D6Xst10 = (2*Vk-Wk) + 2*DFV['6Xs']*Xs - DFV['6Xs']*R
    
    for i in range(1,n-1) :
        h12 = xi[i] - xi[i-1]
        h23 = xi[i+1] - xi[i]
        xi23 = xiplus[i]
        xi12 = ximoins[i]
        upper[i] = -0.5*dt/h23 * (h23*h*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) - 2 * (xi23 * h + Xs) * Dj)  
        DFAupper['1Cm'][i] = -0.5*dt/h23 * (h23*D1Cmh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D1Cmt10 + xi23**2 * (D1Cmh*(Vk-Wk) + h*DFV['1Cm'])  - DFV['1Cm']*Xs- Vk*dt*DFV['1Cm']) - 2 * (xi23 * D1Cmh + dt*DFV['1Cm']) * Dj)  
        DFAupper['1Dj'][i] = -0.5*dt/h23 * (h23*D1Djh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D1Djt10 + xi23**2 * (D1Djh*(Vk-Wk) + h*DFV['1Dj'])  - 1 - DFV['1Dj']*Xs- Vk*dt*DFV['1Dj'])  - 2 * (xi23 * D1Djh + dt*DFV['1Dj']) * Dj- 2 * (xi23 * h + Xs)) 
        DFAupper['1dX'][i] = -0.5*dt/h23 * (h23*D1dXh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D1dXt10 + xi23**2 * (D1dXh*(Vk-Wk) - h*DFW['1dX']))   - 2 * xi23 * D1dXh * Dj) 
        DFAupper['5u'][i]  = -0.5*dt/h23 * (h23*D5uh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D5ut10 + xi23**2 * (D5uh*(Vk-Wk)+ h*DFV['5u'])  - DFV['5u']*Xs - Vk*dt*DFV['5u']) - 2 * (xi23 * D5uh + dt*DFV['5u']) * Dj)  
        DFAupper['6R'][i]  = -0.5*dt/h23 * (h23*D6Rh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D6Rt10 + xi23**2 * D6Rh*(Vk-Wk)+ xi23**2 * h*(DFV['6R']-DFW['6R']) - DFV['6R']*Xs-Vk*D6RXs) - 2 * (xi23 * D6Rh+D6RXs) * Dj)
        DFAupper['6Xs'][i]  = -0.5*dt/h23 * (h23*D6Xsh*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h23*h*(xi23 * D6Xst10 + xi23**2 * D6Xsh*(Vk-Wk) + xi23**2 * h*DFV['6Xs'] - DFV['6Xs']*Xs- Vk*D6XsXs) - 2 * (xi23 * D6Xsh + D6XsXs) * Dj)
        
        lower[i] =  0.5*dt / h12 * (h12 * h * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + 2 * (xi12 * h + Xs) * Dj)
        DFAlower['1Cm'][i] = 0.5*dt / h12 * (h12 * D1Cmh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h12 * h * (xi12 * D1Cmt10 + xi12**2 * D1Cmh*(Vk-Wk) + xi12**2 * h*DFV['1Cm'] - DFV['1Cm']*Xs - Vk*dt*DFV['1Cm']) + 2 * (xi12 * D1Cmh + dt*DFV['1Cm']) * Dj)
        DFAlower['1Dj'][i] = 0.5*dt / h12 * (h12 * D1Djh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h12 * h * (xi12 * D1Djt10 + xi12**2 * D1Djh*(Vk-Wk) + xi12**2 * h*DFV['1Dj']  - 1 - DFV['1Dj']*Xs - Vk*dt*DFV['1Dj']) + 2 * (xi12 * D1Djh + dt*DFV['1Dj']) * Dj + 2 * (xi12 * h + Xs))
        DFAlower['1dX'][i] = 0.5*dt / h12 * (h12 * D1dXh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h12 * h * (xi12 * D1dXt10 + xi12**2 * D1dXh*(Vk-Wk) - xi12**2 * h*DFW['1dX'] )    + 2 * xi12 * D1dXh * Dj)
        DFAlower['5u'][i] = 0.5*dt / h12 * (h12 * D5uh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs)  + h12 * h * (xi12 * D5ut10 + xi12**2 * D5uh*(Vk-Wk)  + xi12**2 * h*DFV['5u']  - DFV['5u']*Xs - Vk*dt*DFV['5u'])   + 2 * (xi12 * D5uh + dt*DFV['5u']) * Dj)
        DFAlower['6R'][i] = 0.5*dt / h12 * (h12 * D6Rh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs)  + h12 * h * (xi12 * D6Rt10 + xi12**2 * D6Rh*(Vk-Wk) + xi12**2 * h*(DFV['6R']-DFW['6R']) - DFV['6R']*Xs - Vk*D6RXs) + 2 * (xi12 * D6Rh+D6RXs) * Dj)
        DFAlower['6Xs'][i] = 0.5*dt / h12 * (h12 * D6Xsh * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + h12 * h * (xi12 * D6Xst10 + xi12**2 * D6Xsh*(Vk-Wk) + xi12**2 * h*DFV['6Xs']  - DFV['6Xs']*Xs- Vk*D6XsXs) + 2 * (xi12 * D6Xsh + D6XsXs) * Dj)
        
        diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
        DFAdiago['1Cm'][i] = 0.5*(((-xi12 + 2 - xi23) * dt*DFV['1Cm'] ) * h**2 * (xi12 - xi23))+ 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1Cmh * (xi12 - xi23)) - DFAupper['1Cm'][i] - DFAlower['1Cm'][i]
        DFAdiago['1Dj'][i] = 0.5*(((-xi12 + 2 - xi23) * dt*DFV['1Dj']) * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1Djh * (xi12 - xi23)) - DFAupper['1Dj'][i] - DFAlower['1Dj'][i] 
        DFAdiago['1dX'][i] = 0.5*(( dt*DFW['1dX'] * (xi12 + xi23)) * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1dXh * (xi12 - xi23))- DFAupper['1dX'][i] - DFAlower['1dX'][i]
        DFAdiago['5u'][i] = 0.5*(((-xi12 + 2 - xi23) * dt*DFV['5u'] ) * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D5uh * (xi12 - xi23)) - DFAupper['5u'][i] - DFAlower['5u'][i]
        DFAdiago['6R'][i] = 0.5*(( (-xi12 + 2 - xi23) * D6RXs +(xi12 + xi23)) * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*D6Rh*h * (xi12 - xi23)) - DFAupper['6R'][i] - DFAlower['6R'][i]
        DFAdiago['6Xs'][i] = 0.5*((-xi12 + 2 - xi23) * D6XsXs  * h**2 * (xi12 - xi23)) + 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D6Xsh * (xi12 - xi23)) - DFAupper['6Xs'][i] - DFAlower['6Xs'][i]
        
        sm[i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))
        DFsm['1Cm'][i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * dt*DFV['1Cm'] ) * h**2 * (xi12 - xi23) + ((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1Cmh * (xi12 - xi23))
        DFsm['1Dj'][i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * dt*DFV['1Dj'] ) * h**2 * (xi12 - xi23) + ((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D1Djh * (xi12 - xi23))
        DFsm['1dX'][i] = ujm[i]*0.5*((dt*DFW['1dX'] * (xi12 + xi23)) * h**2 * (xi12 - xi23) + ((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2 * h*D1dXh * (xi12 - xi23))
        DFsm['2ujm'][i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))
        DFsm['5u'][i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * dt*DFV['5u'] ) * h**2 * (xi12 - xi23) + ((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*D5uh * (xi12 - xi23))
        DFsm['6R'][i] = ujm[i]*0.5*(((-xi12 + 2 - xi23) * D6RXs+D6RR*(xi12 + xi23)) * h**2 * (xi12 - xi23)) + ujm[i]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*D6Rh*h * (xi12 - xi23))
        DFsm['6Xs'][i] = ujm[i]*0.5*((-xi12 + 2 - xi23) * D6XsXs * h**2 * (xi12 - xi23)) + ujm[i]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * D6Xsh*2*h * (xi12 - xi23))
        
    diagonal[n-1] = Dj + (xi[-1]-xi[-2])*(R-Xs)*Wk*(1-kpl)
    DFAdiago['1Cm'][n-1] =  - dt*DFV['1Cm']*(xi[-1]-xi[-2])*Wk*(1-kpl)
    DFAdiago['1Dj'][n-1] =  1 - dt*DFV['1Dj']*(xi[-1]-xi[-2])*Wk*(1-kpl)
    DFAdiago['1dX'][n-1] = (xi[-1]-xi[-2])*dt*DFW['1dX']*Wk*(1-kpl)  + (xi[-1]-xi[-2])*(R-Xs)*DFW['1dX']*(1-kpl)
    DFAdiago['5u'][n-1] =  -dt*DFV['5u']*(xi[-1]-xi[-2])*Wk*(1-kpl)
    DFAdiago['6R'][n-1] = (xi[-1]-xi[-2])*Wk*(1-kpl) + (xi[-1]-xi[-2])*(R-Xs)*DFW['6R']*(1-kpl)
    DFAdiago['6Xs'][n-1] = -(xi[-1]-xi[-2])*D6XsXs*Wk*(1-kpl)
    lower[n-1] = -Dj
    DFAlower['1Dj'][n-1] = -1
    
    return DFV, DFW, DFAupper, DFAlower, DFAdiago, DFsm



def functionF(gamma,double[:] ujm,double[:] u,double R0,double  Xs0,double[:] xi,double dt, constants):
    # calculation of the iteration, whose fixed point = sol. of the implicit time step
    #
    # IN : gamma = dict containing Cm, Dj, dXdt
    #       ujm = u_{j-1} Zr concentration at previous time step
    #       V, W, u : inner/outer velocity, Zr concentration
    #       R / Xs : outer/inner radius
    #       xi : location of space points
    #       dt : time step
    #       constants : a dict containing the constant values
    #
    cdef double x12, x23
    cdef int i
    cdef double Cm = gamma['Cm']
    cdef double Dj = gamma['Dj']
    cdef double dXdt = gamma['dXdt']
    cdef double Czirc = constants['Czirc']
    cdef double kpl = constants['kpl'] 
    cdef int n = np.shape(xi)[0]
    
    cdef double[:] xiplus=np.zeros(n)
    cdef double[:] ximoins=np.zeros(n)
    for i in range(n):
        xiplus[i] = 0.5*xi[i]
        ximoins[i] = 0.5*xi[i]
    for i in range(n-1):
        xiplus[i] += 0.5*xi[i+1]
        ximoins[i+1] += 0.5*xi[i]
    xiplus[n-1] = xi[n-1]+0.5*(xi[n-1]-xi[n-2])
    ximoins[0] = -0.5*xi[1]

    cdef double  Vk = -Dj * (u[1]-Cm)/(xi[1]-xi[0])/(R0-Xs0)/(Cm-Czirc)
    cdef double Wk = -dXdt/3/R0**2
 
    cdef double R = R0+Wk*dt
    cdef double Xs = Xs0+Vk*dt
    cdef double h = R-Xs
    
    cdef double[:] upper = np.zeros(n)
    cdef double[:] lower = np.zeros(n)
    cdef double[:] diagonal = np.zeros(n)
    cdef double[:] sm = np.zeros(n)
    
    diagonal[0] = 1
    sm[0] = Cm
    cdef double t10 = (2*Vk-Wk)*Xs - Vk*R

    for i in range(1,n-1) :
        xi23 = xiplus[i]
        xi12 = ximoins[i]
        
        upper[i] = -0.5*dt * (h*(xi23 * t10 + xi23**2 * h*(Vk-Wk) - Dj - Vk*Xs) - 2 * (xi23 * h + Xs) * Dj/(xi[i+1] - xi[i]))  
        lower[i] =  0.5 * dt * (h * (xi12 * t10 + xi12**2 * h*(Vk-Wk) - Dj - Vk*Xs) + 2 * (xi12 * h + Xs) * Dj/ (xi[i] - xi[i-1]))
        diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
        sm[i] = ujm[i]*0.5*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)
  
    diagonal[n-1] = Dj + (xi[n-1]-xi[n-2])*(R-Xs)*Wk*(1-kpl)
    lower[n-1] = -Dj
    
    return Vk, Wk, upper, lower, diagonal, sm



def functionFTr(double[:] xi,double Vj,double Wj,double[:,:] ujm,double Xs,double R,double dts,double Thj,constants):
    # one iteration in time of the FV scheme, for Tr elements
    # IN : xi : spatials discretization points
    #       Vj/Wj - velocity of the inner/outer boundary
    #       ujm : uTr at the time step j-1
    #       Xs : position of the interface crystal/melt 
    #       R : outer radius of the domain
    #       dts : time step
    #       Thj : temperature at time tj
    #       constants : a dict containing the constant values
    # OUT :    
    #       uTr_j : uTr at the next time step
    #
    # the space variable is  x\in (0,1)
    cdef double ktr, Difk, h12, h23, xi12, xi23
    cdef int i, k
    cdef double xh2o = constants['XH2O']
    cdef int n = np.shape(xi)[0]
    cdef int ntr = np.shape(ujm)[1]
    cdef double[:,:] uTrjk = np.zeros((n,ntr))
    cdef double[:] DifTr = ls.TraceDiff(Thj,xh2o)/constants['Dscale']
    cdef double[:] kt = ls.Kdd(Thj)
    
    cdef double[:] xiplus=np.zeros(n)
    cdef double[:] ximoins=np.zeros(n)
    for i in range(n):
        xiplus[i] = 0.5*xi[i]
        ximoins[i] = 0.5*xi[i]
    for i in range(n-1):
        xiplus[i] += 0.5*xi[i+1]
    xiplus[n-1] = xi[n-1]+0.5*(xi[n-1]-xi[n-2])
    for i in range(1,n):
        ximoins[i] += 0.5*xi[i-1]
    ximoins[0] = -0.5*xi[1]
    
    cdef double h = R-Xs
    cdef double kpl = 0.
    cdef t10 = (2*Vj-Wj)*Xs - Vj*R
    for k in range(ntr):
        ktr = kt[k]
        Difk = DifTr[k]
        
        upper = np.zeros(n)
        lower = np.zeros(n)
        diagonal = np.zeros(n)
        sm = np.zeros(n)
        
        for i in range(1,n-1) :
            h12 = xi[i] - xi[i-1]
            h23 = xi[i+1] - xi[i]
            xi23 = xiplus[i]
            xi12 = ximoins[i]
            
            upper[i] = -0.5*(1/h23 * (h23*h*(xi23 * t10 + xi23**2 * h*(Vj-Wj) - Difk - Vj*Xs) - 2 * (xi23 * h + Xs) * Difk) * dts) 
            lower[i] =  0.5* dts * ( h * (xi12 * t10 + xi12**2 * h*(Vj-Wj) - Difk - Vj*Xs) + 2 * (xi12 * h + Xs) * Difk/ h12)
            diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
            sm[i] = ujm[i,k]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))

        diagonal[0] = Difk - (xi[1]-xi[0])*(R-Xs)*Vj*(1-ktr)
        upper[0] = -Difk
        
        diagonal[n-1] = Difk + (xi[-1]-xi[-2])*(R-Xs)*Wj*(1-kpl)
        lower[n-1] = -Difk
        uTrjkk, tmp = thomas_vecteurs(lower,diagonal,upper,sm)
        for i in range(n):
            uTrjk[i,k] =  uTrjkk[i]
 
    return uTrjk



def derivativeFTr(const double[:] xi,const double Vj,const double Wj,const double DVj,const double DWj,const double[:,:] ujm,const double[:,:] vjm,const double Xs,const double DXs,const double R,const double DR,const double dts,const double Thj,const double alphaj, constants):
    # one iteration in time of the FV scheme, for the derivative of Tr elements

    # IN : xi : spatials discretization points
    #       Vj/Wj - velocity of the inner/outer boundary
    #       DVj/DWj - derivatives of the velocities
    #       ujm : uTr at the time step j-1
    #       vjm : derivative of ujm at time step j-1
    #       Xs : position of the interface crystal/melt
    #       DXs : derivative of Xs
    #       R : outer radius of the domain
    #       DR : derivative of R
    #       dts : time step
    #       Thj : temperature at time tj
    #       alphaj : variation of the temperature at j-th time step   
    #       constants : a dict containing the constant values
    # OUT :    
    #       uTr_j, wTr_j : uTr and its derivative at the next time step
    #
    # the space variable is  x\in (0,1)
 
    cdef int i, k
    cdef double ktr, Difk, Dktr, DDifk, h12, h23, xi12, xi23, t10, Dt10
    cdef double xh2o = constants['XH2O']
    cdef int n = np.shape(xi)[0]
    cdef int ntr = np.shape(ujm)[1]
    cdef double[:,:] uTrjk = np.zeros((n,ntr))
    cdef double[:,:] wTrjk = np.zeros((n,ntr))
    
    cdef double[:] DifTr = ls.TraceDiff(Thj,xh2o)/constants['Dscale']
    cdef double[:] DDifTr = ls.DTraceDiff(Thj,xh2o)/constants['Dscale']*alphaj
    cdef double[:] kt = ls.Kdd(Thj)
    cdef double[:] Dkt = ls.DKdd(Thj)
    for k in range(ntr):
        Dkt[k] = alphaj*Dkt[k]

    cdef double[:] xiplus=np.zeros(n)
    cdef double[:] ximoins=np.zeros(n)
    for i in range(n):
        xiplus[i] = 0.5*xi[i]
        ximoins[i] = 0.5*xi[i]
    for i in range(n-1):
        xiplus[i] += 0.5*xi[i+1]
    xiplus[n-1] = xi[n-1]+0.5*(xi[n-1]-xi[n-2])
    for i in range(1,n):
        ximoins[i] += 0.5*xi[i-1]
    ximoins[0] = -0.5*xi[1]
    
    cdef double h = R - Xs
    cdef double Dh = DR - DXs
    cdef double kpl = 0
    cdef double[:] upper = np.zeros(n)
    cdef double[:] lower = np.zeros(n)
    cdef double[:] diagonal = np.zeros(n)
    cdef double[:] sm = np.zeros(n)
    cdef double[:] Dupper = np.zeros(n)
    cdef double[:] Dlower = np.zeros(n)
    cdef double[:] Ddiagonal = np.zeros(n)
    cdef double[:] Dsm = np.zeros(n)
    cdef double[:] Aprime_u = np.zeros(n)
    cdef double[:] sm2 = np.zeros(n)
    
    for k in range(ntr):
        ktr = kt[k]
        Difk = DifTr[k]
        Dktr = Dkt[k]
        DDifk = DDifTr[k]
    
        t10 = (2*Vj-Wj)*Xs - Vj*R
        Dt10 = (2*DVj-DWj)*Xs + (2*Vj-Wj)*DXs - DVj*R - Vj*DR
        for i in range(1,n-1) :
            h12 = xi[i] - xi[i-1]
            h23 = xi[i+1] - xi[i]
            xi23 = xiplus[i]
            xi12 = ximoins[i]
            
            upper[i] = -0.5* dts/h23 * (h23*h*(xi23 * t10 + xi23**2 * h*(Vj-Wj) - Difk - Vj*Xs) - 2 * (xi23 * h + Xs) * Difk) 
            Dupper[i] = -0.5*dts/h23 * (h23*Dh*(xi23 * t10 + xi23**2 * h*(Vj-Wj) - Difk - Vj*Xs) + h23*h*(xi23 * Dt10 + xi23**2 * Dh*(Vj-Wj)+ xi23**2 * h*(DVj-DWj) - DDifk - DVj*Xs- Vj*DXs) - 2 * (xi23 * Dh + DXs) * Difk- 2 * (xi23 * h + Xs) * DDifk) 
            
            lower[i] =  0.5*(1 / h12 * dts * (h12 * h * (xi12 * t10 + xi12**2 * h*(Vj-Wj) - Difk - Vj*Xs) + 2 * (xi12 * h + Xs) * Difk))
            Dlower[i] =  0.5 / h12 * dts * (h12 * Dh * (xi12 * t10 + xi12**2 * h*(Vj-Wj) - Difk - Vj*Xs)+h12 * h * (xi12 * Dt10 + xi12**2 * Dh*(Vj-Wj)+ xi12**2 * h*(DVj-DWj) - DDifk - DVj*Xs- Vj*DXs) + 2 * (xi12 * Dh + DXs) * Difk+ 2 * (xi12 * h + Xs) * DDifk)
            
            diagonal[i] = 0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23)) - upper[i] - lower[i]
            Ddiagonal[i] = 0.5*((-xi12 + 2 - xi23) * DXs + DR * (xi12 + xi23)) * h**2 * (xi12 - xi23)+0.5*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * 2*h*Dh * (xi12 - xi23) - Dupper[i] - Dlower[i]
                     
            sm[i] = ujm[i,k]*0.5*(((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23))
            Dsm[i] = vjm[i,k]*0.5*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h**2 * (xi12 - xi23) + ujm[i,k]*0.5*((-xi12 + 2 - xi23) * DXs + DR * (xi12 + xi23)) * h**2 * (xi12 - xi23) + ujm[i,k]*((-xi12 + 2 - xi23) * Xs + R * (xi12 + xi23)) * h*Dh * (xi12 - xi23)


        diagonal[0] = Difk - (xi[1]-xi[0])*(R-Xs)*Vj*(1-ktr)
        Ddiagonal[0] = DDifk - (xi[1]-xi[0])*(DR-DXs)*Vj*(1-ktr) - (xi[1]-xi[0])*(R-Xs)*DVj*(1-ktr) + (xi[1]-xi[0])*(R-Xs)*Vj*Dktr
        
        upper[0] = -Difk
        Dupper[0] = -DDifk
        
        diagonal[n-1] = Difk + (xi[-1]-xi[-2])*(R-Xs)*Wj*(1-kpl)
        Ddiagonal[n-1] = DDifk + (xi[-1]-xi[-2])*(DR-DXs)*Wj*(1-kpl)+ (xi[-1]-xi[-2])*(R-Xs)*DWj*(1-kpl)
        
        lower[n-1] = -Difk
        Dlower[n-1] = -DDifk
        
        uTrjkk, tmp =  thomas_vecteurs(lower,diagonal,upper,sm)

        for i in range(n):
            Aprime_u[i] = Ddiagonal[i]*uTrjkk[i]
        for i in range(n-1):
            Aprime_u[i] += Dupper[i]*uTrjkk[i+1]
            Aprime_u[i+1] += Dlower[i+1]*uTrjkk[i]
        for i in range(n):
            sm2[i] = -Aprime_u[i] + Dsm[i]
        wTrjkk, tmp =  thomas_vecteurs(lower,diagonal,upper,sm2)
        for i in range(n):
            uTrjk[i,k] = uTrjkk[i]
            wTrjk[i,k] = wTrjkk[i]
   
    return (uTrjk, wTrjk)

